#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import copy, warnings
from os import path
from collections import OrderedDict
import itertools
import numpy as np
from scipy import stats, signal, interpolate
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
from . import six, afni, io, utils, dicom, math


def convolve_HRF(starts, lens, TR=2, scan_time=None, HRF=None):
    if np.isscalar(lens):
        lens = lens * np.ones(len(starts))
    numout_cmd = '' if scan_time is None else f"-numout {np.ceil(scan_time/TR)}"
    HRF_cmd = '-GAM' if HRF is None else HRF
    res = utils.run(f"waver -TR {TR} {numout_cmd} {HRF_cmd} -tstim {' '.join([f'{t}%{l}' for t, l in zip(starts, lens)])}", verbose=0)
    return np.float_(res['output'])


def create_ideal(stimuli, lens, **kwargs):
    '''
    Parameters
    ----------
    stimuli : list of fname
    '''
    starts = [io.read_stim(fname) for fname in stimuli]
    n_stims = len(starts)
    n_runs = len(starts[0])
    assert(np.all(np.array([len(runs) for runs in starts]) == n_runs))
    if lens == 'alterating':
        lens = [[[] for run in range(n_runs)] for stim in range(n_stims)]
        for run in range(n_runs):
            curr_state, curr_time = -1, np.nan
            n_events = np.array([len(starts[stim][run]) for stim in range(n_stims)])
            ids = np.zeros(n_stims, dtype=int)
            while np.any(ids < n_events):
                wavefront = [(starts[stim][run][ids[stim]] if ids[stim] < n_events[stim] else np.inf) for stim in range(n_stims)]
                next_state, next_time = np.argmin(wavefront), np.min(wavefront)
                if next_state != curr_state:
                    if curr_state != -1:
                        lens[curr_state][run].append(next_time - curr_time)
                    curr_state, curr_time = next_state, next_time
                ids[curr_state] += 1
    elif np.isscalar(lens):
        lens = [[[lens]*len(starts[stim][run]) for run in range(n_runs)] for stim in range(n_stims)]
    ideal = [[convolve_HRF(starts[stim][run], lens[stim][run], **kwargs) for run in range(n_runs)] for stim in range(n_stims)]
    return ideal


def create_times(tmin, tmax, dt):
    if tmin < 0 and tmax > 0:
        times = np.r_[np.arange(-dt, tmin-dt/2, -dt)[::-1], np.arange(0, tmax+dt/2, dt)]
    else:
        times = np.arange(tmin, tmax+dt/2, dt)
    return times


def create_ERP(t, x, events, tmin=-8, tmax=16, dt=0.1, baseline=[-2,0], interp='linear'):
    '''
    t : time for each data point (can be non-contiguous)
    x : [event, feature, time]
    events : event onset time (can be on non-integer time point)
    '''
    times = create_times(tmin, tmax, dt)
    f = interpolate.interp1d(t, x, axis=-1, kind=interp, fill_value=np.nan, bounds_error=False)
    base_corr = create_base_corr_func(times, baseline=baseline)
    ERP = np.zeros(np.r_[len(events), x.shape[1:-1], len(times)].astype(int), dtype=x.dtype)
    for k, t in enumerate(events):
        ERP[k] = base_corr(f(np.arange(t+tmin, t+tmax+dt/2, dt)))
    return ERP, times


class Attributes(object):
    def __init__(self, shape):
        super().__setattr__('attributes', {})
        self.shape = shape

    shape = property(lambda self: self._shape, lambda self, x: setattr(self, '_shape', np.array(x)))

    def add(self, name, value, axis):
        assert(len(value) == self.shape[axis])
        self.attributes[name] = {'axis': axis, 'value': np.array(value)}

    def drop(self, name):
        self.attributes.pop(name)

    def drop_all_with_axis(self, axis):
        axes = axis if np.iterable(axis) else [axis]
        for axis in axes:
            for name in self.names_with_axis(axis):
                self.drop(name)

    def __getattr__(self, name):
        return self.attributes[name]['value']

    def __setattr__(self, name, value):
        if name in self.attributes:
            assert(len(value) == self.shape[self.attributes[name]['axis']])
            self.attributes[name]['value'] = np.array(value)
        else:
            # If the attribute is not added before, it will become a instance attribute
            super().__setattr__(name, value)

    def names_with_axis(self, axis):
        return [name for name, attr in self.attributes.items() if attr['axis']==axis]

    def __repr__(self):
        names = '\n'.join([f" axis={axis} ({self.shape[axis]}) | {', '.join(self.names_with_axis(axis))}" for axis in range(len(self.shape))])
        return f"<Attributes  | shape = {self.shape}\n" + names

    def __copy__(self):
        inst = type(self)(self.shape)
        inst.attributes = copy.deepcopy(self.attributes)
        return inst

    def pick(self, index, axis):
        inst = copy.copy(self)
        if np.iterable(axis):
            indices, axes = index, axis
        else:
            indices, axes = [index], [axis]
        for index, axis in zip(indices, axes):
            # Update shape first via a virtual attribute (in case there is not attribute at all)
            inst.shape[axis] = len(np.arange(inst.shape[axis])[index])
            # Update attributes belonging to the axis
            for name in inst.names_with_axis(axis):
                attr = inst.attributes[name]
                attr['value'] = attr['value'][index]
        return inst

    @classmethod
    def concatinate(cls, attributes_list, axis):
        # Concat shape along axis, and check shape compatibility along other axis
        inst = attributes_list[0]
        self = cls(inst.shape)
        self.shape[axis] = np.sum([attributes.shape[axis] for attributes in attributes_list])
        other_axes = np.r_[0:axis, axis+1:len(self.shape)]
        for attributes in attributes_list[1:]:
            assert(np.all(attributes.shape[other_axes] == self.shape[other_axes]))
        # Concat attributes along axis, and check attributes compatibility (identity) along other axis
        for name, attr in inst.attributes.items():
            if attr['axis'] == axis:
                value = np.concatenate([attributes.attributes[name]['value'] for attributes in attributes_list])
            else:
                value = attr['value']
                for attributes in attributes_list[1:]:
                    assert(np.all(attributes.attributes[name]['value'] == value))
            self.add(name, value, axis)
        return self
                
    def to_dict(self):
        return dict(attributes=self.attributes, shape=self.shape)

    @classmethod
    def from_dict(cls, d):
        self = cls(None)
        for k, v in d.items():
            setattr(self, k, v)
        return self


class Raw(utils.Savable, object):
    def __init__(self, fname, mask=None, TR=None):
        if fname is None:
            return # Skip __init__(), create an empty Raw object, and manually initialize it later.
        if mask is not None:
            self.mask =  mask if isinstance(mask, io.Mask) else io.Mask(mask)
            self.data = self.mask.dump(fname)
        else:
            self.mask = None
            self.data = io.read_vol(fname)
        self.info = {}
        self.info['sfreq'] = 1 / (afni.get_TR(fname) if TR is None else TR)
        self.info['feature_name'] = 'voxel'
        self.info['value_name'] = 'value'
        self.times = np.arange(self.n_times) * self.TR

    shape = property(lambda self: self.data.shape)
    n_features = property(lambda self: self.data.shape[0])
    n_times = property(lambda self: self.data.shape[1])
    TR = property(lambda self: 1 / self.info['sfreq'])

    @classmethod
    def from_array(cls, data, TR):
        '''
        data : 2D array, [n_features, n_times]
        TR : in sec
        '''
        self = cls(None)
        self.mask = None
        self.data = np.array(data, copy=False)
        self.info = dict(sfreq=1/TR, feature_name='voxel', value_name='value')
        self.times = np.arange(self.n_times) * self.TR
        return self
        
    def __repr__(self):
        # return f"<Raw  | {self.n_features} {self.info['feature_name']}s, {self.times[0]:.3f} - {self.times[-1]:.3f} sec, TR = {self.TR} sec, {self.n_times} TRs>"
        return f"<Raw  | {self.n_features} {self.info['feature_name']}s, {self.times[0]:.3f} - {self.times[-1]+self.TR:.3f} sec, TR = {self.TR} sec, {self.n_times} TRs>"

    def copy(self):
        return _copy(self)

    def plot(self, events=None, event_id=None, color=None, palette=None, figsize=None, event_kws=None, **kwargs):
        # Plot mean time course
        data = np.mean(self.data, axis=0)
        if events is not None: # If going to plot events, plot data in black by default
            color = 'k' if color is None else color
        if figsize is not None:
            plt.gcf().set_figwidth(figsize[0])
            plt.gcf().set_figheight(figsize[1])
        plt.plot(self.times, data, color=color, **kwargs)
        plt.xlabel('Time (s)')
        plt.ylabel('Signal change (%)')
        # Plot events
        if events is not None:
            if event_id is None:
                event_id = _default_event_id(events)
            if palette is None: # A palette is eventually a list of colors
                palette = plt.rcParams['axes.prop_cycle'].by_key()['color']
            id2ev = {id: [eid, ev, None] for eid, (ev, id) in enumerate(event_id.items())}
            event_kws = dict(dict(), **(event_kws if event_kws is not None else {}))
            for event in events:
                t, id = event[0], event[-1]
                id2ev[id][2] = plt.axvline(t, color=palette[id2ev[id][0]], **event_kws)
            plt.legend(*zip(*[(h, ev) for eid, ev, h in id2ev.values()]))

    def to_dict(self):
        return dict(info=self.info, data=self.data, mask=self.mask.to_dict(), times=self.times)

    @classmethod
    def from_dict(cls, d):
        self = cls(None)
        for k, v in d.items():
            setattr(self, k, v)
        self.mask = io.Mask.from_dict(self.mask)
        return self


def _copy(self):
    '''Copy all object attributes other than `data`, which is simply referred to.'''
    # TODO: .info and events etc. should be deep copied
    data = self.data
    del self.data
    inst = copy.copy(self)
    inst.data = self.data = data
    return inst


class RawCache(utils.Savable, object):
    def __init__(self, fnames, mask, TR=None, cache_file=None, force_redo=False):
        if fnames is None:
            return # Skip __init__(), create an empty RawCache object, and manually initialize it later.
        if cache_file is None or not utils.exists(cache_file, force_redo=force_redo):
            self.mask = mask if isinstance(mask, io.Mask) else io.Mask(mask)
            self.raws = [Raw(fname, mask=self.mask, TR=TR) for fname in fnames]
            if cache_file is not None:
                self.save(cache_file)
        else:
            inst = self.load(cache_file)
            self.mask = inst.mask
            self.raws = inst.raws

    n_runs = property(lambda self: len(self.raws))

    def get_raws(self, mask, ids=None):
        return_scalar = False
        if ids is None:
            ids = range(self.n_runs)
        elif not utils.iterable(ids):
            return_scalar = True
            ids = [ids]
        if isinstance(mask, six.string_types):
            mask = io.Mask(mask)
            selector = self.mask.infer_selector(mask)
        elif isinstance(mask, io.Mask):
            selector = self.mask.infer_selector(mask)
        else: # boolean index
            selector = mask
            mask = self.mask.pick(selector)
        raws = []
        for idx in ids:
            raw = self.raws[idx].copy()
            raw.data = raw.data[selector]
            raw.mask = mask
            raws.append(raw)
        return raws[0] if return_scalar else raws

    def get_epochs(self, mask, events, event_id, cache_file=None, **kwargs):
        assert(len(events) == self.n_runs)
        if cache_file is None or not utils.exists(cache_file):
            epochs = [Epochs(raw, events[idx], event_id=event_id, **kwargs) for idx, raw in enumerate(self.get_raws(mask))]
            epochs = concatinate_epochs(epochs)
            if cache_file is not None:
                epochs.save(cache_file)
        else:
            epochs = Epochs.load(cache_file)
        return epochs

    def to_dict(self):
        return dict(raws=[raw.to_dict() for raw in self.raws], mask=self.mask.to_dict())

    @classmethod
    def from_dict(cls, d):
        self = cls(None, None)
        for k, v in d.items():
            setattr(self, k, v)
        self.raws = [Raw.from_dict(raw) for raw in self.raws]
        self.mask = io.Mask.from_dict(self.mask)
        return self


def read_events(event_files):
    '''
    Read events from AFNI style (each row is a run, and each element is an occurance) stimulus timing files.

    Parameters
    ----------
    event_files : dict
        e.g., OrderedDict(('Physical/Left', '../stimuli/phy_left.txt'), ('Physical/Right', '../stimuli/phy_right.txt'), ...)

    Returns
    -------
    events_list : list of n_events-by-3 arrays
        The three columns are [start_time(sec), reserved, event_id(int)]
    event_id : dict
    '''
    if not isinstance(event_files, dict):
        event_files = OrderedDict((path.splitext(path.basename(f))[0], f) for f in event_files)
    t = []
    e = []
    event_id = OrderedDict()
    for k, (event, event_file) in enumerate(event_files.items()):
        if isinstance(event_file, str):
            eid = k + 1
        else:
            event_file, eid = event_file
        with open(event_file, 'r') as fi:
            for rid, line in enumerate(fi):
                line = line.strip()
                if not line:
                    continue
                if k == 0:
                    t.append([])
                    e.append([])
                if not line.startswith('*'):
                    t_run = np.float_(line.split())
                    t[rid].extend(t_run)
                    e[rid].extend(np.ones_like(t_run)*(eid))
        event_id[event] = eid
    events_list = []
    for rid in range(len(t)):
        events_run = np.c_[t[rid], np.zeros_like(t[rid]), e[rid]]
        sorter = np.argsort(events_run[:,0])
        events_list.append(events_run[sorter,:])
    return events_list, event_id


def events_from_dataframe(df, run, time, conditions, duration=None, event_id=None):
    if event_id is None:
        levels = itertools.product(*[df[condition].unique() for condition in conditions])
        event_id = OrderedDict([('/'.join(level), k+1) for k, level in enumerate(levels)])
    events = []
    get_event_id = lambda trial: event_id['/'.join([getattr(trial, condition) for condition in conditions])]
    get_duration = lambda trial: 0 if duration is None else getattr(trial, duration)
    for run in df[run].unique():
        events.append(np.array([[trial.time, get_duration(trial), get_event_id(trial)] for trial in df[df.run==run].itertuples()]))
    return events, event_id


def events_to_dataframe(events_list, event_id, conditions):
    id2event = {eid: event.split('/') for event, eid in event_id.items()}
    trials = []
    for rid, events in enumerate(events_list):
        for tid in range(len(events)):
            trials.append(OrderedDict([('run', rid+1), ('trial', tid+1), ('time', events[tid,0])] \
                + list(zip(conditions, id2event[events[tid,2]]))))
    return pd.DataFrame(trials)


def _default_event_id(events):
    return {f"{id:g}": id for id in np.unique(events[:,2])}


def create_base_corr_func(times, baseline=None, method=None):
    '''
    Parameters
    ----------
    time : array-like
        Sampling time for your data.
    baseline : 'none', 'all', or (tmin, tmax)
        Baseline time interval.
    '''
    if baseline is None:
        baseline = 'none'
    if method is None:
        method = np.nanmean
    if isinstance(baseline, str):
        if baseline.lower() == 'none':
            return lambda x: x
        elif baseline.lower() == 'all':
            return lambda x: x - np.nanmean(x, axis=-1, keepdims=True)
    else: # `baseline` is like (tmin, tmax)
        tmin = times[0] if baseline[0] is None else baseline[0]
        tmax = times[-1] if baseline[1] is None else baseline[1]
        times = np.array(times)
        time_sel = (tmin<=times) & (times<=tmax)
        if method in [np.nanmean, np.nanmedian]:
            def base_corr(x):
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', category=RuntimeWarning)
                    return x - method(x[...,time_sel], axis=-1, keepdims=True)
            return base_corr
        else:
            return lambda x: x - method(x[...,time_sel], axis=-1, keepdims=True)


class Epochs(utils.Savable, object):
    def __init__(self, raw, events, event_id=None, tmin=-5, tmax=15, baseline=(-2,0), dt=0.1, interp='linear', hamm=None, conditions=None):
        if raw is None:
            return # Skip __init__(), create an empty Epochs object, and manually initialize it later.
        self.events = events
        self.event_id = _default_event_id(events) if event_id is None else event_id
        self.info = raw.info.copy() # Caution: container datatype is assigned by reference by default
        self.info['sfreq'] = 1 / dt
        self.info['tmin'] = tmin
        self.info['tmax'] = tmax
        self.info['baseline'] = baseline
        self.info['interp'] = interp
        self.info['hamm'] = hamm
        if isinstance(conditions, six.string_types):
            conditions = conditions.split('/')
        self.info['conditions'] = conditions
        self.info['condition'] = None
        self.times = create_times(tmin, tmax, dt)
        x = raw.data
        if hamm is not None:
            h = signal.hamming(hamm)
            h = h/np.sum(h)
            x = signal.filtfilt(h, [1], x, axis=-1)
        f = interpolate.interp1d(raw.times, x, axis=-1, kind=interp, fill_value=np.nan, bounds_error=False)
        base_corr = create_base_corr_func(self.times, baseline=baseline)
        self.data = np.zeros([events.shape[0], raw.n_features, len(self.times)], dtype=raw.data.dtype)
        for k, t in enumerate(events[:,0]):
            self.data[k] = base_corr(f(np.arange(t+tmin, t+tmax+dt/2, dt)))
        self.attr = Attributes(shape=self.shape)

    shape = property(lambda self: self.data.shape)
    n_events = property(lambda self: self.data.shape[0])
    n_features = property(lambda self: self.data.shape[1])
    n_times = property(lambda self: self.data.shape[2])

    @classmethod
    def from_array(cls, data, TR=None, tmin=None, baseline=(-2,0), events=None, event_id=None, conditions=None):
        self = cls(None, None)
        self.data = np.array(data, copy=False)
        assert(self.data.ndim == 3) # n_events x n_features x n_times
        self.info = {}
        self.info['sfreq'] = 1 if TR is None else 1/TR
        self.info['feature_name'] = 'voxel'
        self.info['value_name'] = 'value'
        self.info['tmin'] = 0 if tmin is None else tmin
        self.times = self.info['tmin'] + np.arange(self.data.shape[-1])/self.info['sfreq']
        self.info['tmax'] = self.times[-1]
        self.info['baseline'] = baseline
        self.info['interp'] = None
        self.info['hamm'] = None
        self.info['conditions'] = conditions
        self.info['condition'] = None
        if events is not None:
            assert(self.data.shape[0] == events.shape[0])
            self.events = events
        else:
            self.events = np.zeros([self.data.shape[0], 3])
        if event_id is None:
            self.event_id = {'Event': 0} if events is None else _default_event_id(self.events)
        else:
            assert(np.all(np.in1d(self.events[:,2], list(event_id.values()))))
            self.event_id = event_id
        base_corr = create_base_corr_func(self.times, baseline=baseline)
        self.data = base_corr(self.data)
        self.attr = Attributes(shape=self.shape)
        return self

    def __repr__(self):
        event_str = '\n '.join(f"'{ev}': {np.sum(self.events[:,2]==id)}" for ev, id in self.event_id.items())
        return f"<Epochs  | {self.n_events:4d} events, {self.n_features} {self.info['feature_name']}s, {self.times[0]:.3f} - {self.times[-1]:.3f} sec, baseline {self.info['baseline']}, hamm = {self.info['hamm']},\n {event_str}>"
  
    # https://github.com/mne-tools/mne-python/blob/master/mne/utils/mixin.py
    def __getitem__(self, item):
        if isinstance(item, tuple):
            return self.pick(*item)
        else:
            return self.pick(event=item)

    def add_event_attr(self, name, value):
        self.attr.add(name, value, axis=0)

    def add_feature_attr(self, name, value):
        self.attr.add(name, value, axis=1)

    def pick(self, event=None, feature=None, time=None):
        inst = self.copy()
        # Select event
        if event is None:
            sel_event = slice(None)
        elif isinstance(event, six.string_types) or (utils.iterable(event) and isinstance(event[0], six.string_types)): # e.g., 'Physical/Left'
            sel_event = self._partial_match_event(event)
            self.info['condition'] = event if isinstance(event, six.string_types) else ' | '.join(event) 
        else:
            sel_event = event
        inst.events = inst.events[sel_event]
        inst.event_id = {ev: id for ev, id in inst.event_id.items() if id in inst.events[:,2]}
        # Select feature
        if feature is None:
            sel_feature = slice(None)
        else:
            sel_feature = feature
        # Select time
        if time is None:
            sel_time = slice(None)
        elif utils.iterable(time) and len(time) == 2:
            sel_time = (time[0] <= self.times) & (self.times <= time[1])
        else:
            sel_time = time
        inst.times = inst.times[sel_time]
        inst.info['tmin'] = inst.times[0]
        inst.info['tmax'] = inst.times[-1]
        # inst.info['sfreq'] = ?
        # Make 3D selection
        inst.data = inst.data[sel_event][:,sel_feature][...,sel_time]
        inst.attr = inst.attr.pick([sel_event, sel_feature, sel_time], axis=[0, 1, 2])
        return inst

    def copy(self):
        inst = _copy(self)
        inst.attr = copy.copy(self.attr)
        return inst

    def drop_events(self, ids):
        inst = self.copy()
        inst.data = np.delete(inst.data, ids, axis=0)
        inst.events = np.delete(inst.events, ids, axis=0)
        inst.event_id = {ev: id for ev, id in inst.event_id.items() if id in inst.events[:,2]} # TODO: Need refactor
        inst.attr.pick(np.delete(np.arange(self.n_events), ids), axis=0)
        return inst

    def _partial_match_event(self, keys):
        if isinstance(keys, six.string_types):
            keys = [keys]
        matched = []
        for key in keys:
            key_set = set(key.split('/'))
            matched_id = [id for ev, id in self.event_id.items() if key_set.issubset(ev.split('/'))]
            matched.append(np.atleast_2d(np.in1d(self.events[:,2], matched_id)))
        return np.any(np.vstack(matched), axis=0)

    def apply_baseline(self, baseline):
        base_corr = create_base_corr_func(self.times, baseline=baseline)
        inst = self.copy()
        inst.data = base_corr(inst.data)
        inst.info['baseline'] = baseline
        return inst

    def aggregate(self, event=True, feature=False, time=False, method=np.nanmean, keepdims=np._globals._NoValue, return_index=False):
        axes = ((0,) if event else ()) + ((1,) if feature else ()) + ((2,) if time else ())
        values = method(self.data, axis=axes, keepdims=keepdims)
        events = self.events[:,2] if not event else None
        features = np.arange(self.n_features) if not feature else -1
        times = self.times if not time else np.mean(self.times)
        return (values, events, features, times) if return_index else values

    def average(self, feature=True, time=False, method=np.nanmean, error='bootstrap', ci=95, n_boot=1000, condition=None):
        '''
        Average data over event (and optionally feature and/or time) dimensions, and return an Evoked object.
        '''
        x, _, _, times = self.aggregate(event=False, feature=feature, time=time, method=method, return_index=True)
        data = method(x, axis=0)
        nave = x.shape[0]
        if error == 'bootstrap':
            error_type = (error, ci)
            boot_dist = method(x[np.random.randint(nave, size=[nave, n_boot]),...], axis=0)
            error = np.percentile(boot_dist, [50-ci/2, 50+ci/2], axis=0) - data
        elif error == 'instance':
            error_type = (error, None)
            error = x - data
        if condition is None:
            condition = self.info['condition']
        evoked = Evoked(self.info, data, nave, times, error=error, error_type=error_type, condition=condition)
        return evoked

    # def running_average(self, win_size, overlap=0.5, time=False, method=np.nanmean):
    #     data = []
    #     for start in range(0, self.n_events, int(win_size*overlap)):
    #         x, _, _, times = self.pick(event=slice(start, start+win_size)).aggregate(feature=True, time=time, method=method, return_index=True)
    #         data.append(x)
    #     evoked = Evoked2D(self.info, np.array(data), win_size, features, times)
    
    def transform(self, feature_name, feature_values, transformer):
        '''
        Transform the data into a new Epochs object.
        E.g., epochs.transform('depth', depth, tc.wcutter(depth, linspace(0, 1, 20), win_size=0.2, win_func='gaussian', exclude_outer=True))
        '''
        assert(len(feature_values)==self.n_features)
        inst = self.copy()
        inst.info['feature_name'] = feature_name
        inst.info['feature_values'] = feature_values
        inst.data = transform(inst.data, *transformer, axis=1)
        # TODO: update inst.attr
        return inst


    def summary(self, event=False, feature=True, time=False, method=np.nanmean, attributes=None):
        '''
        Summary data as a pandas DataFrame.
        '''
        assert(self.info['conditions'] is not None)
        dfs = []
        for ev in self.event_id:
            # import pdb; pdb.set_trace()
            x, _, features, times = self.pick(ev).aggregate(event=event, feature=feature, time=time, method=method, keepdims=True, return_index=True)
            events = ev.split('/')
            df = OrderedDict()
            for k, condition in enumerate(self.info['conditions']):
                df[condition] = events[k]
            df[self.info['feature_name']] = np.tile(np.repeat(features, x.shape[2]), x.shape[0])
            df['time'] = np.tile(times, np.prod(x.shape[:2]))
            df[self.info['value_name']] = x.ravel()
            dfs.append(pd.DataFrame(df))
        df = pd.concat(dfs, ignore_index=True)
        if attributes is not None:
            for name, value in attributes.items():
                df[name] = value
        return df

    def plot(self, hue=None, style=None, row=None, col=None, hue_order=None, style_order=None, row_order=None, col_order=None,
        palette=None, dashes=None, figsize=None, bbox_to_anchor=None, subplots_kws=None, average_kws=None, **kwargs):
        assert(self.info['conditions'] is not None)
        conditions = OrderedDict([(condition, np.unique(levels)) for condition, levels in zip(self.info['conditions'], np.array([ev.split('/') for ev in self.event_id]).T)])
        con_sel = [[hue, style, row, col].index(condition) for condition in conditions]
        n_rows = 1 if row is None else len(conditions[row])
        n_cols = 1 if col is None else len(conditions[col])
        subplots_kws = dict(dict(sharex=True, sharey=True, figsize=figsize, constrained_layout=True), **({} if subplots_kws is None else subplots_kws))
        fig, axs = plt.subplots(n_rows, n_cols, squeeze=False, **subplots_kws)
        row_order = [None] if row is None else (conditions[row] if row_order is None else row_order)
        col_order = [None] if col is None else (conditions[col] if col_order is None else col_order)
        hue_order = [None] if hue is None else (conditions[hue] if hue_order is None else hue_order)
        style_order = [None] if style is None else (conditions[style] if style_order is None else style_order)
        if palette is None:
            palette = plt.rcParams['axes.prop_cycle'].by_key()['color']
        elif isinstance(palette, str):
            palette = sns.color_palette(palette)
        if dashes is None:
            dashes = ['-', '--', ':', '-.']
        average_kws = dict(dict(), **({} if average_kws is None else average_kws))
        for rid, row_val in enumerate(row_order):
            for cid, col_val in enumerate(col_order):
                plt.sca(axs[rid,cid])
                show_info = True
                for hid, hue_val in enumerate(hue_order):
                    for sid, style_val in enumerate(style_order):
                        event = '/'.join(np.array([hue_val, style_val, row_val, col_val])[con_sel])
                        label = '/'.join([s for s in [hue_val, style_val] if s is not None])
                        event_sel = self._partial_match_event(event)
                        self[event_sel].average(**average_kws).plot(color=palette[hid], ls=dashes[sid], 
                            label=label, show_n='label' if label else 'info', info=show_info, **kwargs)
                        show_info = False
                        plt.axhline(0, color='gray', ls='--')
                plt.title('/'.join([s for s in [row_val, col_val] if s is not None]))
                if rid < n_rows-1:
                    plt.xlabel('')
                if cid > 0:
                    plt.ylabel('')
                if label:
                    plt.legend(bbox_to_anchor=bbox_to_anchor, loc=None if bbox_to_anchor is None else 'center left')
        sns.despine()

    def to_dict(self):
        return dict(info=self.info, data=self.data, events=self.events, event_id=self.event_id, 
            times=self.times, attr=self.attr.to_dict())

    @classmethod
    def from_dict(cls, d):
        self = cls(None, None)
        for k, v in d.items():
            setattr(self, k, v)
        self.attr = Attributes.from_dict(self.attr)
        return self


def concatinate_epochs(epochs_list, axis=0):
    inst = epochs_list[0].copy()
    inst.data = np.concatenate([epochs.data for epochs in epochs_list], axis=axis)
    inst.attr = Attributes.concatinate([epochs.attr for epochs in epochs_list], axis=axis)
    if axis == 0: # Concat events
        inst.events = np.concatenate([epochs.events for epochs in epochs_list], axis=0)
        inst.event_id = {k: v for epochs in epochs_list for k, v in epochs.event_id.items()}
    elif axis == 1: # Concat voxels
        assert(np.all([np.all(epochs.events == epochs_list[0].events) for epochs in epochs_list[1:]]))
        assert(np.all([(epochs.event_id == epochs_list[0].event_id) for epochs in epochs_list[1:]]))
    return inst


def group_epochs(epochs_list):
    inst = epochs_list[0].copy()
    inst.data = np.concatenate([epochs[event].aggregate(event=True, feature=True, keepdims=True) 
        for epochs in epochs_list for event in inst.event_id], axis=0)
    inst.events = np.array(list(inst.event_id.values())*len(epochs_list))
    inst.events = np.c_[np.zeros((len(inst.events),2)), inst.events]
    inst.attr = Attributes(shape=inst.shape)
    return inst


# def transform(data, idx_gen, weight_gen=None, agg_func=None, axis=0):
#     if agg_func is None:
#         agg_func = np.nanmean if np.any(np.isnan(data)) else np.mean
#     if weight_gen is None:
#         ##### BUG!!!
#         return np.concatenate([agg_func(data[idx,...], axis=axis, keepdims=True) for idx in idx_gen], axis=axis)
#     else:
#         return np.concatenate([agg_func(data[idx,...]*weight, axis=axis, keepdims=True) 
#             for idx, weight in zip(idx_gen, weight_gen)], axis=axis)


def cut(data, val, bins, **kwargs):
    return transform(data, *cutter(val, bins), **kwargs)

def cutter(val, bins):
    idx_gen = ((bins[k]<val)&(val<=bins[k+1]) for k in range(len(bins)-1))
    return idx_gen, None


def qcut(data, val, q, **kwargs):
    return transform(data, *qcutter(val, q), **kwargs)

def qcutter(val, q):
    bins = np.percentile(val, q)
    idx_gen = ((bins[k]<val)&(val<=bins[k+1]) for k in range(len(bins)-1))
    return idx_gen, None


def wcut(data, val, v, win_size, win_func=None, exclude_outer=False, **kwargs):
    return transform(data, *wcutter(val, v, win_size, win_func=win_func, exclude_outer=exclude_outer), **kwargs)

def wcutter(val, v, win_size, win_func=None, exclude_outer=False):
    r = win_size/2
    if win_func == 'gaussian':
        sigma = win_size/4
        win_func = lambda c, x: np.exp(-(x-c)**2/sigma**2)
    if exclude_outer:
        bins = [[max(v[0], vv-r), min(v[-1], vv+r)] for vv in v]
    else:
        bins = [[vv-r, vv+r] for vv in v]
    idx_gen = ((lower<val)&(val<=upper) for lower, upper in bins)
    if win_func is not None:
        weight_gen = (win_func(vv, val[(lower<val)&(val<=upper)]) for vv, (lower, upper) in zip(v, bins))
    else:
        weigth_gen = None
    return idx_gen, weight_gen


class Evoked(object):
    def __init__(self, info, data, nave, times, error=None, error_type=None, condition=None):
        '''
        This class is intended for representing only a single subject and a single condition.
        Group data with multiple conditions together could be flexibly handled by `pd.DataFrame`.
        For the later purpose, instead of Epochs.average() -> Evoked, 
        consider using Epochs.summary() -> pd.DataFrame.
        '''
        self.info = info
        self.times = times
        self.data = data
        self.info['nave'] = nave
        self.error = error
        self.info['error_type'] = error_type
        self.info['condition'] = condition

    shape = property(lambda self: self.data.shape)

    def plot(self, color=None, error=True, info=True, error_kws=None, show_n='info', **kwargs):
        n_str = rf"$n={self.info['nave']}$"
        label = kwargs.pop('label') if 'label' in kwargs else self.info['condition']
        if show_n == 'label':
            label += rf"  ({n_str})"
        line = plt.plot(self.times, self.data, color=color, label=label, **kwargs)[0]
        if color is None:
            color = line.get_color()
        if error:
            if self.info['error_type'][0] == 'instance':
                error_kws = dict(dict(alpha=0.3, lw=0.3), **(error_kws if error_kws is not None else {}))
                plt.plot(self.times, (self.data+self.error).T, color=color, **error_kws)
            else:
                error_kws = dict(dict(alpha=0.3), **(error_kws if error_kws is not None else {}))
                plt.fill_between(self.times, self.data+self.error[0], self.data+self.error[1], color=color, **error_kws)
        plt.xlabel('Time (s)')
        plt.ylabel('Signal change (%)')
        if info:
            ci = self.info['error_type'][1]
            err_str = r'$\pm SEM$' if ci == 68 else rf'${ci}\%\ CI$'
            s = rf"{n_str}$\quad${err_str}" if show_n == 'info' else err_str
            plt.text(0.95, 0.05, s, ha='right', transform=plt.gca().transAxes)


# class Evoked2D(object):
#     def __init__(self, info, data, nave, features, times):
#         self.info = info
#         self.features = features
#         self.times = times
#         self.data = data
#         self.info['nave'] = nave

#     @classmethod
#     def from_evoked_list(cls, evoked_list, features):
#         evoked = evoked_list[0]
#         data = np.array([evoked.data for evoked in evoked_list])
#         return cls(evoked.info, data, evoked.info['nave'], features, evoked.times)


if __name__ == '__main__':
    pass
    # stim_dir = '/Volumes/raid78/ccqian/LaminarRivalry2/LGN/S01/stimuli'
    # stim_files = [f"{stim_dir}/L_replay.txt", f"{stim_dir}/R_replay.txt"]
    # tc.create_ideal(stim_files, 'alterating')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import glob
import numpy as np
from matplotlib import pyplot as plt, transforms
# from matplotlib import transforms
from . import six, io


def get_color_list(cmap):
    '''
    This function is still wrong!
    '''
    r = np.array(cmap._segmentdata['red'])[:,1]
    g = np.array(cmap._segmentdata['green'])[:,1]
    b = np.array(cmap._segmentdata['blue'])[:,1]
    return list(zip(r, g, b))


def draw_color_circle(cmap, res=512):
    '''
    Draw circular colorbar (color index from 0 to 1) clockwise from 0 to 12 o'clock
    '''
    X, Y = np.meshgrid(np.linspace(-1, 1, res), np.linspace(-1, 1, res))
    D = np.sqrt(X**2 + Y**2)
    A = np.rot90((-np.arctan2(-Y, X)+np.pi)/(2*np.pi), 3)
    A[(D<0.4)|(D>1)] = np.nan
    if isinstance(cmap, six.string_types):
        cmap = plt.get_cmap(cmap)
    cmap.set_bad('gray', alpha=1)
    plt.imshow(A, cmap=cmap)


def plot_volreg(dfiles, convention=None):
    '''
    The six columns from left to right are:
    - For 3dvolreg: roll(z), pitch(x), yaw(y), dS(z), dL(x), dP(y)
        Note that the interpretation for roll and yaw in afni is opposite from SPM (and common sense...)
        This program will follow the SPM convention.
    - For 3dAllineate shift_rotate: x-shift  y-shift  z-shift$ z-angle  x-angle$ y-angle$
        Note that the dollar signs in the end indicate parameters that are fixed.
    '''
    files = glob.glob(dfiles) if isinstance(dfiles, six.string_types) else dfiles
    labels = [r'$\Delta$R-L(x) [mm]', r'$\Delta$A-P(y) [mm]', r'$\Delta$I-S(z) [mm]',
              r'Pitch(x) [$^\circ$]', r'Roll(y) [$^\circ$]', r'Yaw(z) [$^\circ$]']
    colors = ['C3', 'C2', 'C0', 'C3', 'C2', 'C0'] # xyz -> RGB
    vrs, comments = zip(*[io.read_txt(f, return_comments=True) for f in files])
    vr = np.vstack(vrs)
    n_runs = len(vrs)
    n_TRs = [vr.shape[0] for vr in vrs]
    borders = np.r_[0, np.cumsum(n_TRs)]
    try:
        base_idx = np.nonzero(np.linalg.norm(vr, axis=1)==0)[0][0]
    except IndexError:
        base_idx = None
    if convention is None:
        if comments[0] and comments[0][0].startswith('# 3dAllineate'):
            convention = '3dAllineate'
        else:
            convention = '3dvolreg'
    if convention == '3dAllineate': # 3dAllineate shift_rotate
        col_order = [0, 1, 2, 4, 5, 3]
        col_sign = np.r_[1, 1, 1, 1, 1, 1]
    elif convention == '3dvolreg': # 3dvolreg
        col_order = [4, 5, 3, 1, 2, 0]
        col_sign = np.r_[-1, -1, -1, 1, 1, 1]
    vr = vr[:,col_order] * col_sign

    fig, axs = plt.subplots(6, 1, sharex=True, sharey=True)
    plt.subplots_adjust(hspace=0, right=0.85)
    for k, ax in enumerate(axs):
        plt.sca(ax)
        plt.plot(vr[:,k], color=colors[k], lw=2)
        plt.axhline(0, color='gray', ls='--', lw=0.5)
        plt.text(1.01, 0.5, labels[k], color=colors[k], transform=ax.transAxes, va='center', fontsize='small')
        if base_idx is not None:
            plt.axvline(base_idx, color='gray', ls='--', lw=0.5)
        for run_idx in range(n_runs):
            if run_idx % 2:
                plt.axvspan(borders[run_idx], borders[run_idx+1], color='C1', alpha=0.2)
    transHDVA = transforms.blended_transform_factory(axs[0].transData, axs[0].transAxes)
    for run_idx in range(n_runs):
        axs[0].text((borders[run_idx]+borders[run_idx+1])/2, 1.05, '{0}'.format(run_idx+1), color='gray', transform=transHDVA, ha='center', fontsize='small')
    plt.sca(axs[-1])
    plt.xlabel('#TR')
    return vr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import unittest
import numpy as np
from mripy import afni, math


class test_afni(unittest.TestCase):
    def test_get_prefix(self):
        # 3d dset
        self.assertEqual(afni.get_prefix('path/to/dset+orig.HEAD'), 'dset')
        self.assertEqual(afni.get_prefix('dset+orig.'), 'dset')
        self.assertEqual(afni.get_prefix('dset+tlrc'), 'dset')
        self.assertEqual(afni.get_prefix('my+dset'), 'my+dset') # Only +orig and +tlrc are treated specially
        self.assertEqual(afni.get_prefix('~/my+dset+orig'), 'my+dset')
        self.assertEqual(afni.get_prefix('path/to/my+dset+orig.BRIK', with_path=True), 'path/to/my+dset')
        # Surface dset
        self.assertEqual(afni.get_prefix('path/to/prefix.niml.dset'), 'prefix')
        self.assertEqual(afni.get_prefix('prefix.1D.dset'), 'prefix')
        self.assertEqual(afni.get_prefix('prefix.niml'), 'prefix')
        self.assertEqual(afni.get_prefix('path/to/prefix.1D'), 'prefix')

    def test_get_suma_spec(self):
        spec = afni.get_suma_spec('path/to/subj_both.spec')
        self.assertEqual(spec['lh'], 'path/to/subj_lh.spec')
        self.assertEqual(sorted(spec.keys()), sorted(afni.SPEC_HEMIS))
        spec = afni.get_suma_spec('subj.hd.lh.spec')
        self.assertEqual(spec['rh'], 'subj.hd.rh.spec')

    def test_substitute_hemi(self):
        self.assertEqual(afni.substitute_hemi('roi.V1.lh.niml.dset'), 'roi.V1.{0}.niml.dset') # In the middle
        self.assertEqual(afni.substitute_hemi('rh.smoothwm.asc'), '{0}.smoothwm.asc') # In the beginning
        self.assertEqual(afni.substitute_hemi('std.60.subj_both.spec', '*'), 'std.60.subj_*.spec') # Flankered by "_"
        self.assertEqual(afni.substitute_hemi('roi.rhymic'), 'roi.rhymic') # Non-standalone instance
        self.assertEqual(afni.substitute_hemi('both_V1.lh.niml.roi'), '{0}_V1.{0}.niml.roi') # No consistancy check for multiple instances

    def test_get_affine(self):
        mat = afni.get_affine(f"testdata/ASR.nii.gz")
        assert(np.allclose(math.apply_affine(mat, np.reshape([0,0,0], [-1,1])), np.reshape([-88,-145,101], [-1,1]), atol=1))
        assert(np.allclose(math.apply_affine(mat, np.reshape([319,319,255], [-1,1])), np.reshape([90,77,-122], [-1,1]), atol=1))
        mat = afni.get_affine(f"testdata/RSA.nii.gz")
        assert(np.allclose(math.apply_affine(mat, np.reshape([0,0,0], [-1,1])), np.reshape([-47,79,16], [-1,1]), atol=1))
        assert(np.allclose(math.apply_affine(mat, np.reshape([319,319,0], [-1,1])), np.reshape([48,79,-79], [-1,1]), atol=1))
        mat = afni.get_affine(f"{res_dir}/ASL.nii.gz")
        assert(np.allclose(math.apply_affine(mat, np.reshape([0,0,0], [-1,1])), np.reshape([46,0,23], [-1,1]), atol=1))
        assert(np.allclose(math.apply_affine(mat, np.reshape([319,319,0], [-1,1])), np.reshape([46,94,-72], [-1,1]), atol=1))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from .. import utils, io, afni, preprocess as prep, surface


def align_suma2exp_mp2rage(suma_dir, T1_al, T1_al_affine=None, T1s=None, T1s_raw=None):
    res_dir, prefix, ext = afni.split_out_file(T1_al, split_path=True, trailing_slash=True)
    if T1_al_affine is None:
        T1_al_affine = f"{res_dir}{prefix}.aff12.1D"
    if T1s is None:
        T1s = f"{res_dir}T1??.nii"
    if T1s_raw is None:
        T1s_raw = f"{res_dir}../raw_fmri/T1??"
    print('>> ========== Create MP2RAGE high SNR region mask ==========')
    prep.assign_mp2rage_labels(T1s, T1s_raw)
    prep.create_mp2rage_SNR_mask(T1s, f"{res_dir}mp2rage_mask.nii")
    prep.apply_transforms(T1_al_affine, T1_al, \
        f"{res_dir}mp2rage_mask.nii", f"{res_dir}mp2rage_mask_al.nii", interp='NN')
    print('>> ========== ANTs align SurfVol to T1_al ==========')
    surf_vol = afni.get_surf_vol(suma_dir)
    outputs = prep.align_ants(T1_al, surf_vol, f"{res_dir}SurfVol_al.nii", 
        strip=[prep.copy_dset, prep.copy_dset], base_mask=f"{res_dir}mp2rage_mask_al.nii")
    print('>> ========== Transform SUMA surface meshes and volumes ==========')
    surface.transform_suma(outputs['transform'], suma_dir)
    suma_subj = afni.get_suma_subj(suma_dir)
    prep.create_suma_script(f"{suma_dir+'.al'}/{suma_subj}_both.spec", \
        f"{res_dir}SurfVol_al.nii", f"{res_dir}run_suma_al")


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import argparse, textwrap, glob, re, sys, os, shutil
from os import path
import numpy as np


def plot_match(physio_infos, series_infos):
    from matplotlib import pyplot as plt
    from matplotlib import transforms

    transHDVA = transforms.blended_transform_factory(plt.gca().transData, plt.gca().transAxes)
    n_plots = 0
    appended = False
    for k, sinfo in enumerate(series_infos):
        plt.axvspan(sinfo['start'], sinfo['stop'], n_plots*0.2, n_plots*0.2+0.2, color='k', alpha=0.3)
        plt.text(sinfo['start'], n_plots*0.2+0.05, '%d'%k, transform=transHDVA, clip_on=True)
        print('DICOM#{0:02d}: {1} ({2} volumes)'.format(k, path.basename(sinfo['files'][0]), sinfo['n_volumes']))
        appended = True
    if appended:
        plt.text(0.01, n_plots*0.2+0.1, 'DICOM', transform=plt.gca().transAxes)
        print('='*30)
        n_plots += 1
    for ch, color in zip(['resp', 'puls'], ['b', 'r']):
        appended = False
        for k, pinfo in enumerate(physio_infos):
            if pinfo is not None and ch in pinfo:
                plt.axvspan(pinfo[ch]['start'], pinfo[ch]['stop'], n_plots*0.2, n_plots*0.2+0.2, color=color, alpha=0.3)
                plt.text(pinfo[ch]['start'], n_plots*0.2+0.05, '%d'%k, transform=transHDVA, clip_on=True)
                print('{0}#{1:02d}: {2} ({3} samples)'.format(ch.upper(), k, path.basename(pinfo[ch]['file']), len(pinfo[ch]['data'])))
                appended = True
        if appended:
            plt.text(0.01, n_plots*0.2+0.1, ch.upper(), transform=plt.gca().transAxes)
            print('='*30)
            n_plots += 1
    plt.show()


if __name__ == '__main__':
    import script_utils # Append mripy to Python path
    from mripy import io, utils, six
    print = six.print_

    parser = argparse.ArgumentParser(description='Extract physiological data sync with mri acquisition.',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=textwrap.dedent('''\
            examples:
              1) Physiological files are in physio/, and raw dicom files are in
                 func01/, func02/, func03/, etc. Simply list timing information.
                $ extract_physio.py -p physio -d func* -l
              2) Both physiological and dicom files are in 20170222_S18. Valid
                 functional runs are from 0006 to 0009. Output extracted 1D files
                 to 20170222_S18/extracted.
                $ extract_physio.py -p 20170222_S18 -d 20170222_S18 -f 6-9 -o\
            '''))
    parser.add_argument('-p', '--physio', default='.', metavar='path/to/physio/dir', help='folder that contains *.resp/*.puls/etc.')
    parser.add_argument('-d', '--dicom', default=['.'], nargs='+', metavar='path/to/dicom/dir', help='one or more folders containing *.IMA')
    parser.add_argument('-o', '--output_dir', nargs='?', const='default', metavar='path/to/output/dir', help='folder to output the 1D files')
    parser.add_argument('-l', '--list', action='store_true', help='list timing info')
    parser.add_argument('-C', '--channels', default=['resp', 'puls'], nargs='+', help='channels to list (resp/puls/etc.)')
    parser.add_argument('-f', '--func', default=[], nargs='+', help='series number of selected func runs')
    parser.add_argument('-u', '--dummy', default=0, help='number of dummy scans (will trim physio accordingly)')
    parser.add_argument('-c', '--copy', nargs='?', const='default', metavar='path/to/copy/files/to', help='copy raw physio files to specified folder')
    parser.add_argument('-nbf', '--name_by_folder', action='store_true', help='name output by dicom folder names')
    parser.add_argument('-g', '--graph', action='store_true', help='plot match between physio and dicom')
    parser.add_argument('-m', '--match', default='cover', help='matching method: (cover)|overlap')
    parser.add_argument('-M', '--CMRR', action='store_true', help='CMRR MB acquisition time correction')
    args = parser.parse_args()
    if args.output_dir == 'default':
        args.output_dir = path.join(args.physio, 'physio')
    if args.output_dir is not None and not path.exists(args.output_dir):
        os.makedirs(args.output_dir)
    if args.copy == 'default':
        folder_name = 'raw_physio'
        if args.output_dir is not None:
            args.copy = path.join(path.split(args.output_dir)[0], folder_name)
        else:
            args.copy = path.join(args.physio, folder_name)
    if args.copy is not None and not path.exists(args.copy):
        os.makedirs(args.copy)
    args.func = utils.expand_index_list(args.func)
    # Parse dicom files
    print('Parsing dicom info...', end='', flush=True)
    dicom_files = []
    for dicom_dir in args.dicom:
        dicom_files.extend(io.filter_dicom_files(dicom_dir, series_numbers=args.func, instance_numbers=[1]))
    series_infos = [io.parse_series_info(f, shift_time='CMRR' if args.CMRR else None) for f in dicom_files]
    print(' ({0} files)'.format(len(series_infos)))
    if len(series_infos) == 0:
        print('No dicom file found in {0}. Use -d to specify one or more dicom input dir(s).'
            .format(', '.join(['"{0}"'.format(path.realpath(p)) for p in args.dicom])))
        exit()
    # Parse physiological files
    print('Parsing physiological info...', end='', flush=True)
    physio_files = sorted(glob.glob(path.join(args.physio, '*.resp')))
    date = series_infos[0]['date'] # Potential bug: Assume there is no cross-day experiment
    physio_infos = [io.parse_physio_files(f, date=date, channels=args.channels) for f in physio_files]
    print(' ({0} files)'.format(len(physio_infos)))
    if len(physio_infos) == 0:
        print('No physiological file found in "{0}". Use -p to specify physiological input dir.'
            .format(path.realpath(args.physio)))
        exit()
    # Match physiological with
    if args.graph:
        plot_match(physio_infos, series_infos)
        # from mypy import mypy
        # mypy.saveObjectAsPickle(dict(physio_infos=physio_infos, series_infos=series_infos), 'debug.pickle')
    physio_infos, series_infos = io.match_physio_with_series(physio_infos, series_infos,
        channel=args.channels[0], method=args.match)
    print('{0} pairs were found.'.format(len(physio_infos)))
    # List timing
    if args.list:
        print('==================================================')
        for k, sinfo in enumerate(series_infos):
            for ch in args.channels:
                io._print_physio_timing(physio_infos[k][ch], sinfo, ch, k)
    # Extract 1D files
    if args.output_dir is not None:
        print('Writing outputs...')
        for k, sinfo in enumerate(series_infos):
            res = io.extract_physio(physio_infos[k], sinfo, TR=sinfo['TR'], dummy=int(args.dummy), channels=args.channels, verbose=int(not args.list))
            dicom_folder_name = path.split(path.split(sinfo['files'][0])[0])[1]
            for c, ch in enumerate(args.channels):
                if args.name_by_folder:
                    fname = '{0}_{1}.1D'.format(ch, dicom_folder_name)
                else:
                    fname = '{0}{1:02d}.1D'.format(ch, k+1)
                np.savetxt(path.join(args.output_dir, fname), res[c], fmt=str('%d')) # Use str() because npyio.py is not 2/3 compatible
    # Copy raw physio files
    if args.copy is not None:
        print('Copying raw physio files...')
        for physio_info in physio_infos:
            for f in glob.glob(path.splitext(list(physio_info.values())[0]['file'])[0] + '.*'):
                shutil.copy(f, args.copy)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import glob, re, itertools, inspect
import gzip
from datetime import datetime
from collections import OrderedDict
from os import path
import numpy as np
import six


encoding = 'utf-8'

vr_parsers = {
    'CS': lambda x: str.strip(x.decode(encoding)), # Code String
    'SH': lambda x: str.strip(x.decode(encoding)), # Short String
    'LO': lambda x: str.strip(x.decode(encoding)), # Long String
    'ST': lambda x: str.strip(x.decode(encoding)), # Short Text
    'LT': lambda x: str.strip(x.decode(encoding)), # Long Text
    'UT': lambda x: str.strip(x.decode(encoding)), # Unlimited Text

    'PN': lambda x: str.strip(x.decode(encoding)), # Person Name
    'AS': lambda x: str.strip(x.decode(encoding)), # Age String ??? e.g., 039Y

    'DA': lambda x: datetime.strptime(x.decode(encoding), '%Y%m%d').date(), # Date
    'TM': lambda x: datetime.strptime(x.decode(encoding).strip(), '%H%M%S.%f').time(), # Time

    'IS': lambda x: np.int_(x.decode(encoding).split('\\')).squeeze()[()], # Integer String
    'DS': lambda x: np.float_(x.decode(encoding).split('\\')).squeeze()[()], # Decimal String

    'SS': lambda x: np.frombuffer(x, dtype=np.int16).squeeze()[()], # Signed Short
    'US': lambda x: np.frombuffer(x, dtype=np.uint16).squeeze()[()], # Unsigned Short
    'SL': lambda x: np.frombuffer(x, dtype=np.int32).squeeze()[()], # Signed Long
    'UL': lambda x: np.frombuffer(x, dtype=np.uint32).squeeze()[()], # Unsigned Long
    'FL': lambda x: np.frombuffer(x, dtype=np.float32).squeeze()[()], # Floating Point Single
    'FD': lambda x: np.frombuffer(x, dtype=np.float64).squeeze()[()], # Floating Point Double
}

# `gdcmdump --csa input.dcm`
# `dcmstack` in Python
# https://nipy.org/nibabel/dicom/siemens_csa.html
# https://neurostars.org/t/determining-bids-phaseencodingdirection-from-dicom/612 # For PhaseEncodingDirection
def parse_Siemens_CSA(b):
    return {}

def parse_Siemens_CSA2(b):
    def parse_value(pattern, b, default=None):
        match = re.search(pattern, b)        
        return match.group(1) if match else default
    CSA2 = {
        'ReferenceAmplitude': float(parse_value(rb'sTXSPEC\.asNucleusInfo\[0\]\.flReferenceAmplitude\s+=\s+(\S+)\s', b, default=np.nan)),
        'PhasePartialFourier': re.search(rb'sKSpace\.ucPhasePartialFourier\s+=\s+(\S+)\s', b).group(1).decode(encoding),
        'SlicePartialFourier': re.search(rb'sKSpace\.ucSlicePartialFourier\s+=\s+(\S+)\s', b).group(1).decode(encoding),
        'RefLinesPE': int(parse_value(rb'sPat\.lRefLinesPE\s+=\s+(\S+)\s', b, default=0)),
        'PATMode': re.search(rb'sPat.ucPATMode\s+=\s+(\S+)\s', b).group(1).decode(encoding),
        'RefScanMode': re.search(rb'sPat\.ucRefScanMode\s+=\s+(\S+)\s', b).group(1).decode(encoding),
        'TotalScanTimeSec': float(re.search(rb'lTotalScanTimeSec\s+=\s+(\S+)\s', b).group(1)),
        # 'SlicePosition': np.float_(re.findall(rb'sSliceArray\.asSlice\[\d+\]\.sPosition\.\w{4}\s+=\s(\S+)\s', b)).reshape(-1,3), # Unfortunately, no temporal order
    }
    return CSA2

Siemens = {
    'PartialFourier': {
        '0x10': None,
        '0x8': '7/8',
        '0x4': '6/8',
        '0x2': '5/8',
    },
    'PATMode': {
        '0x1': None, #??
        '0x2': 'GRAPPA',
    },
    'RefScanMode': {
        '0x1': None, #??
        '0x4': 'GRE',
    },
}


# http://dicom.nema.org/medical/dicom/current/output/chtml/part06/chapter_6.html
# https://dicom.innolitics.com/ciods
# Note that the hex codes must be in upper case. 
tag_parsers = {
    '0010,0010': ('PatientName', vr_parsers['PN']),
    '0010,0030': ('PatientBirthDate', vr_parsers['DA']),
    '0010,0040': ('PatientSex', vr_parsers['CS']),
    '0010,1010': ('PatientAge', vr_parsers['AS']),
    '0010,1030': ('PatientWeight', vr_parsers['DS']),
    '0002,0013': ('ImplementationVersionName', vr_parsers['SH']),
    '0008,0022': ('AcquisitionDate', vr_parsers['DA']),
    '0008,0032': ('AcquisitionTime', vr_parsers['TM']),
    '0008,103E': ('SeriesDescription', vr_parsers['LO']),
    '0018,0020': ('ScanningSequence', vr_parsers['CS']), # SE Spin Echo, IR Inversion Recovery, GR Gradient Recalled, EP Echo Planar, RM Research Mode
    '0018,0021': ('SequenceVariant', vr_parsers['CS']), # SK segmented k-space, MTC magnetization transfer contrast, SS steady state, TRSS time reversed steady state, SP spoiled, MP MAG prepared, OSP oversampling phase, NONE no sequence variant
    '0018,0023': ('MRAcquisitionType', vr_parsers['CS']), # 2D frequency x phase, 3D frequency x phase x phase
    '0018,0024': ('SequenceName', vr_parsers['SH']), # User defined name for the combination of Scanning Sequence (0018,0020) and Sequence Variant (0018,0021)
    '0018,0050': ('SliceThickness', vr_parsers['DS']),
    '0018,0080': ('RepetitionTime', vr_parsers['DS']),
    '0018,0081': ('EchoTime', vr_parsers['DS']),
    '0018,0082': ('InversionTime', vr_parsers['DS']),
    '0018,0084': ('ImagingFrequency', vr_parsers['DS']),
    # '0018,0086': ('EchoNumber', vr_parsers['IS']), # The echo number used in generating this image. In the case of segmented k-space, it is the effective Echo Number. (However, could be 1 for MB GE EPI)
    '0018,0087': ('MagneticFieldStrength', vr_parsers['DS']),
    '0018,0088': ('SpacingBetweenSlices', vr_parsers['DS']), # Thickness plus gap
    # '0018,0091': ('EchoTrainLength', vr_parsers['IS']), # Number of lines in k-space acquired per excitation per image. (However, could be 1 for MB GE EPI)
    '0018,0095': ('PixelBandwidth', vr_parsers['DS']),
    '0018,1020': ('SoftwareVersion', vr_parsers['LO']),
    '0018,1030': ('ProtocolName', vr_parsers['LO']),
    '0018,1251': ('TransmittingCoil', vr_parsers['SH']),
    '0018,1310': ('AcquisitionMatrix', lambda x : np.array([a for a in vr_parsers['US'](x) if a != 0])), # Dimensions of the acquired frequency/phase data before reconstruction: frequency rows\frequency columns\phase rows\phase columns
    '0018,1312': ('PhaseEncodingDirection', vr_parsers['CS']),
    '0018,1314': ('FlipAngle', vr_parsers['DS']),
    '0018,1316': ('SAR', vr_parsers['DS']),
    '0019,100A': ('n_slices', vr_parsers['US']),
    # https://wiki.humanconnectome.org/download/attachments/40534057/CMRR_MB_Slice_Order.pdf
    # https://wiki.humanconnectome.org/download/attachments/40534057/CMRR_MB_Slice_Order.pdf?version=2&modificationDate=1386950067494&api=v2
    # About Siemens slice timing: "For odd, the most inferior slice is acquired first. For even, the most inferior slice is acquired second."
    # However, "slice excitation always starts with slice0 in CMRR multiband C2P sequences" (i.e., different from Siemens default behavior).
    # "Slice cross-talk effects are minimized with interleaved slice series, hence this is the selected option in the default protocol."
    # "The most convenient and practical way to determine slice timing is by referencing the timing information for each slice under "MosaicRefAcqTimes" 
    # ([ms], ordered corresponding to the slice numbering) in vendor private field of the DICOM header. 
    # This slice-by-slice timing information is generic (transparent to the multiband factor) for any protocol."
    '0019,1029': ('MosaicRefAcqTimes', vr_parsers['FD']),
    '0020,0010': ('StudyID', lambda x : int(vr_parsers['SH'](x))),
    '0020,0011': ('SeriesNumber', vr_parsers['IS']),
    '0020,0012': ('AcquisitionNumber', vr_parsers['IS']),
    '0020,0013': ('InstanceNumber', vr_parsers['IS']),
    '0020,4000': ('ImageComments', vr_parsers['LT']),
    '0028,0030': ('PixelSpacing', vr_parsers['DS']),
    '0029,1010': ('CSA', parse_Siemens_CSA), # Siemens private element
    '0029,1020': ('CSA2', parse_Siemens_CSA2), # Siemens private element
}

Siemens_parsers = {
    '0051,100E': ('slice_orientation', vr_parsers['SH']),
    '0051,1011': ('acceleration_factor', vr_parsers['SH']),
    '0051,1016': ('reconstruction', vr_parsers['SH']),
}

custom_parsers = {
    'resolution': lambda header: np.r_[header['PixelSpacing'], header['SliceThickness']],
    'FOV': lambda header: header['resolution'][:2] * header['AcquisitionMatrix'], # Bug: Is "PixelSpacing" also ordered as frequency/phase like "AcquisitionMatrix" does??
    'orientation': lambda header: ('oblique-' if len(header['slice_orientation'])>3 else '') + {'Sag': 'sagittal', 'Cor': 'coronal', 'Tra': 'transversal'}[header['slice_orientation'][:3]],
    'GRAPPA': lambda header: int(header['acceleration_factor'].split()[0][1:]) if 'acceleration_factor' in header and header['acceleration_factor'].startswith('p') else 0, # TODO: The text can be "p2" or "p2 s4". What does "s4" mean?
    'PhasePartialFourier': lambda header: Siemens['PartialFourier'][header['CSA2']['PhasePartialFourier']],
    'SlicePartialFourier': lambda header: Siemens['PartialFourier'][header['CSA2']['SlicePartialFourier']],
    'MultiBand': lambda header: int(re.search('MB(\d+)', header['ImageComments']).group(1)) if 'MB' in header['ImageComments'] else None, # https://github.com/CMRR-C2P/MB/issues/223
    'distortion_correction': lambda header: re.search('(ND|DIS2D|DIS3D)', header['reconstruction']).group(1),
    'ReferenceAmplitude': lambda header: header['CSA2']['ReferenceAmplitude'],
    'PATMode': lambda header: Siemens['PATMode'][header['CSA2']['PATMode']],
    'RefLinesPE': lambda header: header['CSA2']['RefLinesPE'],
    'RefScanMode': lambda header: Siemens['RefScanMode'][header['CSA2']['RefScanMode']],
    'TotalScanTime': lambda header: header['CSA2']['TotalScanTimeSec'],
    'timestamp': lambda header: datetime.combine(header['AcquisitionDate'], header['AcquisitionTime']).timestamp(), 
}


def parse_SQ_data_element(fi):
    '''
    We only support Data Element with Explicit VR at present (Table 7.1-2).
    We don't support nested Item at present (2018-09-26).

    References
    ----------
    [1] http://dicom.nema.org/Dicom/2013/output/chtml/part05/chapter_7.html
    '''
    while True:
        # Parse an item
        item_tag = '{0:04X},{1:04X}'.format(*np.frombuffer(fi.read(4), dtype=np.uint16, count=2))
        if item_tag == 'FFFE,E000': # Item (Mark the start of an item)
            item_length = int.from_bytes(fi.read(4), byteorder='little', signed=False)
            if item_length == 4294967295: # 0xFFFFFFFF: Undefined Length
                # Bruteforce scan the byte stream until we hit "FFFE,E00D"
                # Bug: However, this may fail if the payload also contains item with undefined length!
                while True:
                    # Until we hit "FFFE,E00D"
                    if fi.read(2) == b'\xfe\xff':
                        if fi.read(2) == b'\x0d\xe0':
                            if fi.read(4) == b'\x00\x00\x00\x00':
                                break
            else:
                fi.seek(item_length, 1)
        elif item_tag == 'FFFE,E00D': # Item Delimitation Item (Mark the end of an item with undefined length)
            item_length = int.from_bytes(fi.read(4), byteorder='little', signed=False)
            assert(item_length == 0)
        elif item_tag == 'FFFE,E0DD': # Sequence Delimitation Item (Mark the end of an SQ with undefined length)
            item_length = int.from_bytes(fi.read(4), byteorder='little', signed=False)
            assert(item_length == 0)
            break


def parse_dicom_header(fname, search_for_tags=None, **kwargs):
    '''
    Parameters
    ----------
    fname : str
    search_for_tags : set
        Search for specific dicom tags, and stop file scanning early if all tags of interest are seen.
        e.g., search_for_tags={'0020,0011', '0020,0013'} will search for SeriesNumber and InstanceNumber.
        This will save you some time, esp. when the remote file is accessed via slow data link.
    **kwargs : 
        This is only for backward compatibility.

    Notes
    -----
    "Implicit and Explicit VR Data Elements shall not coexist in a Data Set and Data Sets nested within it 
    (see Section 7.5). Whether a Data Set uses Explicit or Implicit VR, among other characteristics, 
    is determined by the negotiated Transfer Syntax (see Section 10 and Annex A)." [1]

    References
    ----------
    [1] http://dicom.nema.org/Dicom/2013/output/chtml/part05/chapter_7.html
    [2] https://stackoverflow.com/questions/119684/parse-dicom-files-in-native-python
    '''
    elements = []
    header = OrderedDict()
    n_tags_seen = 0
    tag_parsers.update(Siemens_parsers)
    if fname.endswith('.gz'):
        opener = gzip.open
    else:
        opener = open
    with opener(fname, 'rb') as fi:
        # The preamble
        fi.seek(128) # The first 128 bytes are 0x00
        assert(fi.read(4).decode(encoding) == 'DICM') # The next 4 bytes are "DICM"
        # Data Elements
        while True:
            element = {}
            group = fi.read(2)
            if not group:
                break
            element['group'] = '{0:04X}'.format(int.from_bytes(group, byteorder='little', signed=False))
            element['element'] = '{0:04X}'.format(int.from_bytes(fi.read(2), byteorder='little', signed=False))
            tag = ','.join((element['group'], element['element']))
            # print(tag, end='')
            element['VR'] = fi.read(2).decode(encoding)
            # print(':', element['VR'])
            if element['VR'] in ['OB', 'OW', 'OF', 'SQ', 'UT', 'UN']:
                fi.seek(2, 1)
                element['length'] = int.from_bytes(fi.read(4), byteorder='little', signed=False)
            else:
                element['length'] = int.from_bytes(fi.read(2), byteorder='little', signed=False)
            if element['length'] == 4294967295: # 0xFFFFFFFF: Undefined Length
                if element['VR'] == 'SQ':
                    parse_SQ_data_element(fi)
                else:
                    # print(element['VR'])
                    raise NotImplementedError('** Undefined Length')
            elif tag in tag_parsers:
                header[tag_parsers[tag][0]] = tag_parsers[tag][1](fi.read(element['length']))
            else:
                fi.seek(element['length'], 1)
            elements.append(element)
            if search_for_tags is not None and tag in search_for_tags:
                n_tags_seen += 1
                if n_tags_seen == len(search_for_tags):
                    break
    # elements = pd.DataFrame(elements, columns=['group', 'element', 'VR', 'length'])
    # Custom header fields
    for field, parser in custom_parsers.items():
        try:
            header[field] = parser(header)
        except KeyError:
            pass
    header['filename'] = path.realpath(fname)
    return header


def sort_dicom_series(folder):
    '''
    Parameters
    ----------
    folder : string
        Path to the folder containing all the dicom files.

    Returns
    -------
    studies : list of dicts
        [{'0001': [file0, file1, ...], '0002': [files], ...}, {study1}, ...]
    '''
    exts = ['.IMA', '.dcm', '.dcm.gz']
    files = sorted(itertools.chain.from_iterable(glob.glob(path.join(folder, '*'+ext)) for ext in exts))
    headers = [parse_dicom_header(f, search_for_tags={'0020,0010', '0020,0011', '0020,0013'}) for f in files]
    studies = []
    for study_id in np.unique([header['StudyID'] for header in headers]):
        study = OrderedDict()
        for series_id in np.unique([header['SeriesNumber'] for header in headers if header['StudyID'] == study_id]):
            series = sorted([(path.basename(header['filename']), header['InstanceNumber']) for header in headers \
                if header['StudyID']==study_id and header['SeriesNumber']==series_id], key=lambda x: x[1])
            study['{0:04d}'.format(series_id)] = [x[0] for x in series]
        studies.append(study)
    return studies


def parse_series_info(dicom_files, dicom_ext=None, parser=None, return_headers=False):
    '''
    Parameters
    ----------
    dicom_files : list or str
        A list of dicom files (e.g., as provided by sort_dicom_series), or
        a folder that contains a single series (e.g., "../raw_fmri/func01"), or 
        a single dicom file.
    '''
    if dicom_ext is None:
        dicom_ext = '.IMA'
    if parser is None:
        parser = parse_dicom_header
    if isinstance(dicom_files, six.string_types): # A single file or a folder
        if path.isdir(dicom_files):
            # Assume there is only one series in the folder
            dicom_files = sorted(glob.glob(path.join(dicom_files, '*'+dicom_ext)))
        else:
            dicom_files = [dicom_files]
    # Parse dicom headers
    headers = [parser(f) for f in dicom_files]
    info = OrderedDict(headers[0])
    assert(np.all(np.array([header['StudyID'] for header in headers])==info['StudyID']))
    assert(np.all(np.array([header['SeriesNumber'] for header in headers])==info['SeriesNumber']))
    # n_volumes, n_slices, TR
    info['n_volumes'] = headers[-1]['AcquisitionNumber'] - headers[0]['AcquisitionNumber'] + 1
    if 'n_slices' not in info:
        info['n_slices'] = int(len(headers) / info['n_volumes'])
    info['first'] = headers[0]['timestamp']
    info['last'] = headers[-1]['timestamp']
    info['TR'] = (info['last']-info['first'])/(info['n_volumes']-1) if info['n_volumes'] > 1 else None
    # Slice timing
    # if shift_time == 'CMRR':
    #     shift_time = 0
    #     if info['TR'] is not None and 'n_slices' in info and np.mod(info['n_slices'], 2)==0:
    #         slice_order = pares_slice_order(files)[0]
    #         if slice_order == 'interleaved':
    #             shift_time = -info['TR']/2
    # elif shift_time is None:
    #     shift_time = 0
    # info['first'] += shift_time
    # info['last'] += shift_time
    info['start'] = info['first']
    info['stop'] = (info['last'] + info['TR']) if info['TR'] is not None else info['last']
    info['time'] = np.array([header['timestamp'] for header in headers]) - info['timestamp']
    if return_headers:
        info['headers'] = headers
    return info



if __name__ == '__main__':
    print(parse_dicom_header('20180626_S18_EP2DBR_S07.MR.S18_APPROVED.0010.0001.2018.06.26.12.47.59.31250.120791562.IMA'))
    print(parse_dicom_header('20180918_S18_FACEID_VIS_S01.MR.S18_APPROVED.0007.0001.2018.09.18.15.52.59.828125.140479.IMA'))
    pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import argparse, textwrap, subprocess, glob, re, os
from os import path

# 1. Clusterize
# https://afni.nimh.nih.gov/afni/community/board/read.php?1,137288,137299
# You can't drive the clusterize plugin directly yet...
# 2. Layout
# `afni -layout -` is required to suppress the default layout (open axi/sag/cor image)
# 3. DriveSuma
# `-switch_surf lh.inflated` is required to switch state

def parse_window_mini_language(desc):
    info = {}
    for item in desc.split():
        key, value = item.split('=')
        if key == 'geom':
            match = re.match('(?:(.+)x(.+))?([+|-].+)([+|-].+)', value)
            if match:
                if match.group(1) is not None:
                    info['w'] = int(match.group(1))
                    info['h'] = int(match.group(2))
                info['x'] = int(match.group(3))
                info['y'] = int(match.group(4))
            match = re.match('(.+)x(.+)', value)
            if match:
                info['w'] = int(match.group(1))
                info['h'] = int(match.group(2))
        else:
            info[key] = value
    return info


if __name__ == '__main__':
    import script_utils # Append mripy to Python path
    from mripy import afni

    CONTROLLER_HEIGHT = 412
    WINDOW_OFFSET = 426

    parser = argparse.ArgumentParser(description='Open afni/suma to view results with least possible clicks.',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=textwrap.dedent('''\
            examples:
              1) ULay = SurfVol_Alnd_Exp+orig
                 OLay = stats.*REML+orig'[RetinoL-R#0_Coef]'
                 Thr = stats.*REML+orig'[RetinoL-R#0_Tstat]'
                 p = 0.05, clim = 1,
                 view axialimage and sagittalgraph, move Xhairs to I J K
                 press t to link suma.
                $ afni_viewer.py -l RetinoL-R -a -S -ijk 82 180 167 --link
              2) ULay = T1+orig
                 OLay = stats+orig'[L-R#0_Tstat]'
                 Thr = stats+orig'[L+R#0_Tstat]'
                 view all three image, no suma.
                $ afni_viewer.py -u T1+orig -o stats+orig -v L-R#t -t L+R#t -suma off\
            '''))
    parser.add_argument('-u', '--ulay', default='infer', help='underlay dset, default SurfVol_Alnd_Exp+orig > T1_al+orig')
    parser.add_argument('-o', '--olay', default='infer', help='overlay dset, default stats.*REML+orig')
    parser.add_argument('-l', '--label', help='sub-brick label for OLay/Thr')
    parser.add_argument('-v', '--value', default='infer', help='sub-brick index/label for OLay')
    parser.add_argument('-t', '--threshold', default='infer', help='sub-brick index/label for Thr')
    parser.add_argument('-p', default='0.05', help='p value corresponding to overlay threshold')
    parser.add_argument('-r', '--range', default='1', help='color range')
    parser.add_argument('-m', '--resample', default='NN', help='resampling mode for OLay/Thr, default NN: {NN|Li|Cu|Bk}[.{NN|Li|Cu|Bk}]')
    parser.add_argument('-no', '--olay_off', action='store_true', help='hide overlay')
    parser.add_argument('-a', '--axialimage', nargs='?', const='', help='show axial image: -a geom=800x500+900+0 crop=260:330,240:310')
    parser.add_argument('-s', '--sagittalimage', nargs='?', const='', help='show sagittal image')
    parser.add_argument('-c', '--coronalimage', nargs='?', const='', help='show coronal image')
    parser.add_argument('-A', '--axialgraph', nargs='?', const='', help='show axial graph')
    parser.add_argument('-S', '--sagittalgraph', nargs='?', const='', help='show sagittal graph')
    parser.add_argument('-C', '--coronalgraph', nargs='?', const='', help='show coronal graph')
    parser.add_argument('-ijk', '--ijk', nargs=3, metavar=('I', 'J', 'K'), help='crosshairs coordinates in voxel indices')
    parser.add_argument('-xyz', '--xyz', nargs=3, metavar=('X', 'Y', 'Z'), help='crosshairs coordinates in DICOM order')
    parser.add_argument('-suma', '--suma', nargs='?', const='', help='show suma, with "geom=800x500+900+0 state=inflated"')
    parser.add_argument('--suma_dir', default='../SUMA', help='path to SUMA dir')
    parser.add_argument('--spec', default='', help='spec file')
    parser.add_argument('--surf_vol', default='infer', help='spec file')
    parser.add_argument('--viewer_cont', default='', help='additional viewer control sub-commands. --viewer_cont="..."')
    parser.add_argument('--camera', default='-key:r7 up -key:r36 right', help='rotate suma camera with arrow keys')
    parser.add_argument('-link', '--link', action='store_const', const='-key t', default='', help='link suma with afni with t key')
    parser.add_argument('--surf_cont', default='', help='additional surface control sub-commands. --surf_cont="..."')
    parser.add_argument('--controller', action='store_true', help='show object controller')
    parser.add_argument('-d', '--dset', help='surface dset')
    parser.add_argument('--dry', action='store_true', help='print cmd without actual execution')
    args = parser.parse_args()

    if args.ulay == 'infer':
        if path.exists('SurfVol_Alnd_Exp+orig.HEAD'):
            args.ulay = 'SurfVol_Alnd_Exp+orig'
            if args.suma is None:
                args.suma = ''
        elif path.exists('T1_al+orig.HEAD'):
            args.ulay = 'T1_al+orig'
        else:
            raise ValueException('-u ULAY is required.')
    if args.olay == 'infer':
        dsets = glob.glob('stats.*REML+orig.HEAD')
        if dsets:
            args.olay = dsets[0]
        elif args.olay_off:
            args.olay = args.ulay
        else:
            raise ValueException('-o OLAY is required.')
    label2index = afni.get_brick_labels(args.olay, label2index=True)
    n_bricks = len(label2index)
    if args.value == 'infer':
        if args.label is not None:
            args.value = args.label + '#0_Coef'
        else:
            args.value = '{0}'.format(n_bricks-3) # Presumably beta value
    args.value = re.sub('#b$', '#0_Coef', args.value)
    if args.value in label2index:
        args.value = label2index[args.value]
    if args.threshold == 'infer':
        if args.label is not None:
            args.threshold = args.label + '#0_Tstat'
        else:
            args.threshold = '{0}'.format(n_bricks-2) # Presumably t value
    args.threshold = re.sub('#t', '#0_Tstat', args.threshold)
    if args.threshold in label2index:
        args.threshold = label2index[args.threshold]
    if (args.axialimage is None and args.sagittalimage is None and args.coronalimage is None and
        args.axialgraph is None and args.sagittalgraph is None and args.coronalgraph is None):
        args.axialimage = ''
        args.sagittalimage = ''
        args.coronalimage = ''

    # print(args)
    environ = {'AFNI_NOSPLASH': 'YES'}
    if args.olay_off:
        olay_cmd = 'SET_FUNC_VISIBLE A.-'
    else:
        olay_cmd = 'SET_FUNC_VISIBLE A.+; \
            OPEN_PANEL A.Define_Overlay; \
            SET_OVERLAY A.{olay} {value} {threshold}; \
            SET_THRESHNEW A {p} *p; \
            SET_FUNC_RANGE A.{range}; \
            SET_FUNC_RESAM A.{resample}'.format(
            olay=afni.get_prefix(args.olay),
            value=args.value,
            threshold=args.threshold,
            p=args.p,
            range=args.range,
            resample=args.resample)
    win_names = ['axialimage', 'sagittalimage', 'coronalimage', 'axialgraph', 'sagittalgraph', 'coronalgraph']
    win_cmds = []
    win_x = 0
    for name in win_names:
        if args.__getattribute__(name) is not None:
            cmd = 'OPEN_WINDOW A.{0} {1}'.format(name, args.__getattribute__(name))
            if 'geom' not in cmd:
                cmd += 'geom=+{0}+{1}'.format(win_x, CONTROLLER_HEIGHT)
            win_x += WINDOW_OFFSET
            win_cmds.append(cmd)
    win_cmd = '; '.join(win_cmds)
    if args.xyz is not None:
        xyz_cmd = 'SET_DICOM_XYZ A {0}'.format(' '.join(args.xyz))
    elif args.ijk is None:
        xyz_cmd = ''
    elif args.ijk == ['q', 'c', 'c']:
        xyz_cmd = ''
    else:
        xyz_cmd = 'SET_IJK A {0}'.format(' '.join(args.ijk))
    # `afni -layout -` is required to suppress the default layout (open axi/sag/cor image)
    afni_cmd = 'afni -layout - -niml -com "\
        {win_cmd}; \
        SET_UNDERLAY A.{ulay} 0; \
        {olay_cmd}; \
        {xyz_cmd}" &'.format(
        ulay=afni.get_prefix(args.ulay),
        olay_cmd=olay_cmd,
        win_cmd=win_cmd,
        xyz_cmd=xyz_cmd)
    print(afni_cmd)
    if not args.dry:
        subprocess.Popen(afni_cmd, env=dict(os.environ, **environ), shell=True) # subprocess.call(cmd, shell=True)

    if args.suma is not None and args.suma.lower() not in ['none', 'off'] :
        if not args.spec or not path.exists(path.join(args.suma_dir, args.spec)):
            suma_subj = afni.get_suma_subj(args.suma_dir)
            args.spec = '{0}/{2}{1}_both.spec'.format(args.suma_dir, suma_subj, args.spec)
        if args.surf_vol == 'infer':
            args.surf_vol = args.ulay
        suma_info = dict(w=800, h=500, x=900, y=0, state='inflated')
        suma_info.update(parse_window_mini_language(args.suma))
        if not re.search('\b[lrm]h\b', suma_info['state']):
            suma_info['state'] = 'lh.' + suma_info['state'] # Only `-switch_surf lh.inflated`
        if not args.dset is None:
            args.surf_cont = '-load_dset {0} '.format(args.dset) + args.surf_cont
            args.controller = True
        if args.controller:
            args.surf_cont = '-view_object_cont y ' + args.surf_cont
        suma_cmd = 'suma \
            -spec {spec_file} \
            -sv {surf_vol} \
            -drive_com "\
            -com viewer_cont -viewer_size {viewer_size} -viewer_position {viewer_pos} \
            {viewer_cont} {camera} {link} \
            -com surf_cont -switch_surf {state} {surf_cont} \
            " &'.format(
            spec_file=args.spec,
            surf_vol=args.surf_vol,
            viewer_size='{0} {1}'.format(suma_info['w'], suma_info['h']),
            viewer_pos='{0} {1}'.format(suma_info['x'], suma_info['y']),
            state=suma_info['state'],
            viewer_cont=args.viewer_cont,
            camera=args.camera,
            link=args.link,
            surf_cont=args.surf_cont)
        # suma -spec ../SUMA/YangY_both.spec -sv SurfVol_Alnd_Exp+orig -drive_com "-com viewer_cont -key:r10 right" &
        print(suma_cmd)
        if not args.dry:
            subprocess.call(suma_cmd, shell=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import time
import numpy as np
from scipy import optimize, stats
from deepdish import io as dio
from . import utils, math


def basis_vanBergen2015(s, n_channels=8, power=5):
    '''
    Parameters
    ----------
    s : 1D array, n_trials
    
    Returns
    -------
    fs : 2D array, n_channels * n_trials

    References
    ----------
    {van Bergen2015}
    '''
    phi = np.arange(0, np.pi, np.pi/n_channels)
    fs = np.maximum(0, np.cos(2 * (s[np.newaxis,:] - phi[:,np.newaxis])))**power # n_channels * n_trials
    return fs


def basis_Sprague2013(s, n_channels=6, spacing=2, center=None, size=None, power=7, dim=1, intercept=False):
    '''
    Parameters
    ----------
    s : 2D array, n_trials * dim | 1D array, n_trials
    
    Returns
    -------
    fs : 2D array, n_channels * n_trials

    References
    ----------
    {Sprague2013}
    '''
    if size is None:
        size = 5.8153/2.0940 * spacing # Ratio is chosen to avoid high corr between channels while accomplish smooth recon
    if center is None:
        center = 0
    centers = (np.arange(n_channels) - (n_channels-1)/2) * spacing + center
    if dim == 1:
        r = np.abs(s[np.newaxis,:] - centers[:,np.newaxis]) # Distance from filter’s center
    elif dim == 2:
        X, Y = np.meshgrid(centers, centers)
        centers = np.c_[X.ravel(), Y.ravel()] # 2D channel array is serialized in row-first order
        r = np.linalg.norm(s.T[np.newaxis,...] - centers[...,np.newaxis], axis=1)
    fs = np.where(r<size, (0.5*np.cos(r/size*np.pi) + 0.5)**power, 0) # n_channels * n_trials
    if intercept:
        fs = np.vstack([fs, np.ones(len(s))]) # Learnable bias
    return fs


class BaseModel(utils.Savable2):
    # get_params() is required by sklearn
    def get_params(self, deep=True):
        raise NotImplementedError('** You must implement this method by yourself in your model.') # Abstrast method

    # set_params() is required by sklearn
    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self

    def to_dict(self):
        return {k: v for k, v in self.__dict__.items() if k.endswith('_')}

    def from_dict(self, d):
        self.__dict__.update(d)
        return self


class ChannelEncodingModel(BaseModel):
    def __init__(self, n_channels, basis_func, stimulus_domain, circular=False, verbose=2):
        '''
        After (Brouwer et al., 2009).
        '''
        self.n_channels = n_channels
        self.basis_func = basis_func
        self.stimulus_domain = stimulus_domain
        self.circular = circular
        self.verbose = verbose

    # get_params() is required by sklearn
    def get_params(self, deep=True):
        return dict(n_channels=self.n_channels, basis_func=self.basis_func, 
            stimulus_domain=self.stimulus_domain, circular=self.circular)

    def fit(self, X, y):
        '''
        Parameters
        ----------
        X : 2D array
            n_trials * n_voxels BOLD response pattern
            (e.g., beta for each trial, or delayed and detrended time points within block plateau)
        y : 1D array
            n_trials stimulus value (e.g., orientation, color)
        '''
        b = X.T # Voxel BOLD response, n_voxels * n_trials
        s = y # Stimulus, n_trials
        fs = self.basis_func(s) # Channel response, n_channels * n_trials
        # Step 1: Estimate W by OLS regression
        W = b @ fs.T @ math.pinv(fs @ fs.T) # Weight, n_voxels * n_channels
        # Store params
        self.W_ = W
        return self # Required by sklearn

    def predict(self, X, stimulus_domain=None, return_all=False):
        stimulus_domain = self.stimulus_domain if stimulus_domain is None else stimulus_domain
        channel_resp = self.inverted_encoding(X) # n_trials * n_channels
        evidence = self.correlation_inversion(X, stimulus_domain=None) # n_trials * n_domain
        # y_map = stimulus_domain[np.argmax(evidence, axis=1)] # n_trials
        y_map = stimulus_domain[math.median_argmax(evidence, axis=1)] # n_trials
        # y_mean = np.sum(stimulus_domain * evidence, axis=1) / np.sum(evidence, axis=1) # n_trials
        # y_std = np.sqrt(np.sum((stimulus_domain[np.newaxis,:] - y_mean[:,np.newaxis])**2 * evidence, axis=1) \
        #     / np.sum(evidence, axis=1)) # n_trials
        # return (y_mean, y_std, y_map, evidence, channel_resp) if return_all else y_mean
        return (y_map, evidence, channel_resp) if return_all else y_map

    _pidx = 1 # Index of posterior (for MAP) if return_all

    def pRF(self, stimulus_domain=None, method='ols', X=None, y=None):
        stimulus_domain = self.stimulus_domain if stimulus_domain is None else stimulus_domain
        fs_domain = self.basis_func(stimulus_domain) # n_channels * n_domain
        if method == 'ols':
            pRF = self.W_ @ fs_domain # n_voxels * n_domain
        elif method == 'ridge': 
            # After {Sprague2013}, to ensure that most of the pRFs were sufficiently unimodal
            # However, the method doesn't seem to help much in my simulation...
            b = X.T # Voxel BOLD response, n_voxels * n_trials
            s = y # Stimulus, n_trials
            fs = self.basis_func(s) # Channel response, n_channels * n_trials
            n_trials = len(s) 
            # Find best lambda according to BIC
            Ws = []
            BICs = []
            for lambda_ in np.logspace(-3, 3, 19):
                M = fs.T @ math.pinv(fs @ fs.T + lambda_ * np.eye(self.n_channels)) # n_trials * n_channels
                W = b @ M # Weight, n_voxels * n_channels
                H = M @ fs # Projection (or "hat") matrix, n_trials * n_trials
                df = np.trace(H) # Degrees of freedom for ridge (less than the number of predictors as in OLS)
                RSS = np.sum((b @ H - b)**2) # Residual sum of squares
                BIC = n_trials * np.log(RSS) + df * np.log(n_trials) # Scaling problem???
                Ws.append(W)
                BICs.append(BIC)
            W_opt = Ws[np.argmin(BICs)]
            pRF = W_opt @ fs_domain
        return pRF # n_voxels * n_domain
        
    def inverted_encoding(self, X):
        b = X.T # Voxel BOLD response, n_voxels * n_trials
        fs = math.pinv(self.W_.T @ self.W_) @ self.W_.T @ b # Inverted channel response, n_channels * n_trials
        return fs.T # n_trials * n_channels

    def voxel_inversion(self, X, stimulus_domain=None):
        stimulus_domain = self.stimulus_domain if stimulus_domain is None else stimulus_domain
        recon = X @ self.pRF(stimulus_domain=stimulus_domain) # n_trials * n_domain
        recon /= X.shape[1] # / n_voxels
        return recon

    def channel_inversion(self, X, stimulus_domain=None):
        stimulus_domain = self.stimulus_domain if stimulus_domain is None else stimulus_domain
        fs_domain = self.basis_func(stimulus_domain) # n_channels * n_domain
        fs_data = self.inverted_encoding(X) # n_trials * n_channels
        recon = fs_data @ fs_domain # n_trials * n_domain
        return recon

    def correlation_inversion(self, X, stimulus_domain=None):
        stimulus_domain = self.stimulus_domain if stimulus_domain is None else stimulus_domain
        # Iterate every possible stimulus in the domain
        fs_domain = self.basis_func(stimulus_domain) # n_channels * n_domain
        # Inverted encoding
        fs_data = self.inverted_encoding(X) # n_trials * n_channels
        # Correlation as evidence (similar to likelihood)
        n_trials, n_domain = fs_data.shape[0], fs_domain.shape[1]
        evidence = np.zeros([n_trials, n_domain]) # n_trials * n_domain
        for tid in range(n_trials):
            for did in range(n_domain):
                evidence[tid,did] = np.corrcoef(fs_data[tid,:], fs_domain[:,did])[0,1]
        return evidence


class BayesianChannelModel(BaseModel):
    def __init__(self, n_channels='required', basis_func='required', stimulus_domain='required', circular=False, stimulus_prior=None, global_search=False, verbose=2):
        '''
        After (van Bergen et al., 2015).

        Examples
        --------
        from mripy import encoding
        from sklearn import model_selection, preprocessing, pipeline

        stimulus_domain = np.linspace(0, pi, 181)
        n_channels = 8
        basis_func = lambda s: encoding.basis_vanBergen2015(s, n_channels=n_channels)
        model = encoding.BayesianChannelModel(n_channels=n_channels, 
            basis_func=basis_func, stimulus_domain=stimulus_domain, circular=True)
        model = pipeline.make_pipeline(preprocessing.StandardScaler(), model)
        cv = model_selection.LeaveOneGroupOut() # One group of each run
        y_hat = model_selection.cross_val_predict(model, X, y, groups, cv=cv, n_jobs=1)
        '''
        self.n_channels = n_channels
        self.basis_func = basis_func
        self.stimulus_domain = stimulus_domain
        self.circular = circular
        self.stimulus_prior = 1 if stimulus_prior is None else stimulus_prior # TODO: This should be refactored for CV
        self.global_search = global_search
        self.verbose = verbose

    # get_params() is required by sklearn
    def get_params(self, deep=True):
        return dict(n_channels=self.n_channels, basis_func=self.basis_func, stimulus_domain=self.stimulus_domain,
            circular=self.circular, stimulus_prior=self.stimulus_prior, global_search=self.global_search)

    def fit(self, X, y):
        '''
        Parameters
        ----------
        X : 2D array
            n_trials * n_voxels BOLD response pattern
            (e.g., beta for each trial, or delayed and detrended time points within block plateau)
        y : 1D array
            n_trials stimulus value (e.g., orientation, color)
        '''
        b = X.T # Voxel BOLD response, n_voxels * n_trials
        s = y # Stimulus, n_trials
        fs = self.basis_func(s) # Channel response, n_channels * n_trials
        # Step 1: Estimate W by OLS regression
        # W = b @ fs.T @ np.linalg.inv(fs @ fs.T) # Weight, n_voxels * n_channels
        # It is critical to use pinv() here instead of inv():
        # 1) Result in much more accurate estimation for W, and hence tau, rho, sigma
        # 2) Avoid "LinAlgError: Singular matrix" at `inv(fs @ fs.T)`
        # 3) Avoid negative negloglikelihood (i.e., negative `slogdet(Omega)`) 
        #    and non positive semidefinite `Omega` (i.e., `all(eigvals(Omega)>0) == False`), 
        #    which could occur with randn W
        # Note that Gilles used `np.linalg.lstsq()` here, which should be numerically adept.
        W = b @ fs.T @ math.pinv(fs @ fs.T) # Weight, n_voxels * n_channels
        # Store params
        self.W_ = W
        # Step 2: Estimate tau, rho, sigma by ML optimization (gradient-based)
        z = b - W @ fs # n_voxels * n_trials
        # Initial params
        tau0 = np.std(b, axis=1)
        rho0 = np.mean(np.corrcoef(b)[np.triu_indices(len(tau0), k=1)])
        sigma0 = np.mean(np.std(fs, axis=1))
        params0 = np.r_[tau0, rho0, sigma0/5.0]
        # bounds = np.c_[np.ones(len(params0))*1e-4, np.r_[tau0*5, 1, sigma0*5]]
        bounds = np.c_[np.ones(len(params0))*1e-3, np.r_[tau0*5, 0.99, sigma0*5]]
        # Conjugate gradient algorithm, due to lack of support for bounds, requires multi-start to avoid/alleviate being trapped in local minima.
        if self.verbose > 0:
            print('>> Start maximum likelihood optimization...')
        if self.global_search:
            def accept_test(f_new, x_new, f_old, x_old):
                Omega = self._calc_Omega(self.W_, x_new[:-2], x_new[-2], x_new[-1])
                # is_pos_semi_def = np.all(np.linalg.eigvals(Omega) > 0)
                # return (f_new < f_old and f_new > 0 and is_pos_semi_def)
                is_singular = np.linalg.matrix_rank(Omega, hermitian=True) < Omega.shape[0]
                return (f_new < f_old and f_new > 0 and not is_singular)
            res = optimize.basinhopping(self._negloglikelihood, params0, accept_test=accept_test, 
                minimizer_kwargs=dict(args=(z, W), method='L-BFGS-B', jac=self._negloglikelihood_prime, bounds=bounds))
        else:
            class Counter(object):
                def __init__(self, model, args):
                    self.count = 0
                    self.last_time = time.time()
                    self.model = model
                    self.args = args
                def step(self, xk):
                    self.count += 1
                    # cost = self.model._negloglikelihood(xk, *self.args) # The pinv->svd here is too expensive...
                    curr_time = time.time()
                    duration = curr_time - self.last_time
                    self.last_time = curr_time
                    if self.model.verbose > 1:
                        # print(f"iter#{self.count:03d} ({utils.format_duration(duration)}): cost={cost:.4f}, tau[-3:]={xk[-5:-2]}, rho={xk[-2]:.4f}, sigma={xk[-1]:.4f}")
                        print(f"iter#{self.count:03d} ({utils.format_duration(duration)}): tau[-3:]={xk[-5:-2]}, rho={xk[-2]:.4f}, sigma={xk[-1]:.4f}")
                    elif self.model.verbose == 1:
                        print(f"iter#{self.count:03d} ({utils.format_duration(duration)}): tau[-3:]={xk[-5:-2]}, rho={xk[-2]:.4f}, sigma={xk[-1]:.4f}", end='\r')
            # res = optimize.minimize(self._negloglikelihood, params0, args=(z, W), method='L-BFGS-B', 
            #     jac=self._negloglikelihood_prime, bounds=bounds, callback=Counter(model=self, args=(z, W)).step)
            res = optimize.minimize(self._negloglikelihood, params0, args=(z, W, True), method='L-BFGS-B', 
                jac=True, bounds=bounds, callback=Counter(model=self, args=(z, W)).step)
        params = res.x
        if self.verbose > 0:
            print(f"cost={res.fun}, iter={res.nit}, func_eval={res.nfev}, success={res.success}, {res.message}")
            print(params0[-3:], '-->', params[-3:])
        # Store params
        self.tau_, self.rho_, self.sigma_ = params[:-2], params[-2], params[-1]
        self._Omega = self._calc_Omega(self.W_, self.tau_, self.rho_, self.sigma_)
        self._Omega_inv = math.pinv(self.Omega_) # Update cache
        return self # Required by sklearn

    # Cache backed properties (the "if else" construct is to prevent unnecessary expression evaluation)
    Omega_ = property(lambda self: self.__dict__.setdefault('_Omega', None if hasattr(self, '_Omega') else self._calc_Omega(self.W_, self.tau_, self.rho_, self.sigma_)))
    Omega_inv_ = property(lambda self: self.__dict__.setdefault('_Omega_inv', None if hasattr(self, '_Omega_inv') else math.pinv(self.Omega_)))

    def predict(self, X, stimulus_domain=None, stimulus_prior=None, return_all=False):
        stimulus_domain = self.stimulus_domain if stimulus_domain is None else stimulus_domain
        stimulus_prior = self.stimulus_prior if stimulus_prior is None else stimulus_prior
        posterior = self.bayesian_inversion(X, stimulus_domain=stimulus_domain, stimulus_prior=stimulus_prior) # n_trials * n_domain
        y_map = stimulus_domain[np.argmax(posterior, axis=1)] # n_trials
        # y_map = stimulus_domain[math.median_argmax(posterior, axis=1)] # n_trials
        if self.circular:
            y_mean = math.circular_mean(stimulus_domain, domain=stimulus_domain, weight=posterior, axis=1)
            y_std = math.circular_std(stimulus_domain, domain=stimulus_domain, weight=posterior, axis=1)
        else:
            y_mean = np.sum(stimulus_domain * posterior, axis=1) / np.sum(posterior, axis=1) # n_trials
            y_std = np.sqrt(np.sum((stimulus_domain[np.newaxis,:] - y_mean[:,np.newaxis])**2 * posterior, axis=1) \
                / np.sum(posterior, axis=1)) # n_trials
        return (y_mean, y_std, y_map, posterior) if return_all else y_mean

    _pidx = 3 # Index of posterior (for MAP) if return_all

    def loglikelihood(self, X, y):
        b = X.T
        s = y
        fs = self.basis_func(s)
        z = b - self.W_ @ fs
        return self._calc_L(z, self.W_, self.tau_, self.rho_, self.sigma_)

    def bayesian_inversion(self, X, stimulus_domain=None, stimulus_prior=None, density=True):
        '''
        Parameters
        ----------
        X : 2D array, n_trials * n_voxels
        stimulus_domain : 1D array, n_domain
        stimulus_prior : 2D array, n_trials * n_domain (or 1D array, n_domain)
            None for a flat stimulus prior, same for all trials.

        Returns
        -------
        posterior : 2D array, n_trials * n_domain
        '''
        stimulus_domain = self.stimulus_domain if stimulus_domain is None else stimulus_domain
        stimulus_prior = self.stimulus_prior if stimulus_prior is None else stimulus_prior
        b = X.T[:,np.newaxis,:] # n_voxels * n_domain * n_trials
        fs = self.basis_func(stimulus_domain) # n_channels * n_domain
        predicted_mean_resp = (self.W_ @ fs)[...,np.newaxis] # n_voxels * n_domain * n_trials
        z = b - predicted_mean_resp
        # # mv_norm = stats.multivariate_normal(np.zeros(self.Omega_.shape[0]), self.Omega_)
        # mv_norm = stats.multivariate_normal(np.zeros(self.Omega_.shape[0]), self.Omega_, allow_singular=True)
        # # likelihood = mv_norm.pdf(z.T) # n_trials * n_domain
        # # posterior = likelihood * stimulus_prior # n_trials * n_domain
        # # posterior /= np.sum(posterior, axis=1, keepdims=True)
        # # The above code will underflow
        # loglikelihood = mv_norm.logpdf(z.T) # n_trials * n_domain
        # The following implementation is 10x faster than stats.multivariate_normal:D
        loglikelihood = math.gaussian_logpdf(z.T, np.zeros(self.Omega_.shape[0]), self.Omega_, cov_inv=self.Omega_inv_) # n_trials * n_domain
        logposterior = loglikelihood + np.log(stimulus_prior) # n_trials * n_domain
        posterior = math.normalize_logP(logposterior, axis=1)
        if density:
            posterior /= stimulus_domain[-1] - stimulus_domain[0]
        return posterior

    def _negloglikelihood(self, params, z, W, return_prime=False):
        tau, rho, sigma = params[:-2], params[-2], params[-1]
        if not return_prime:
            return -self._calc_L(z, W, tau, rho, sigma)
        else:
            Omega = self._calc_Omega(W, tau, rho, sigma)
            Omega_inv = math.pinv(Omega) # np.linalg.pinv() may encounter "LinAlgError: SVD did not converge" for some matrices
            L = -self._calc_L(z, W, tau, rho, sigma, Omega=Omega, Omega_inv=Omega_inv)
            L_prime = self._negloglikelihood_prime(params, z, W, Omega=Omega, Omega_inv=Omega_inv)
            return L, L_prime

    def _negloglikelihood_prime(self, params, z, W, Omega=None, Omega_inv=None):
        tau, rho, sigma = params[:-2], params[-2], params[-1]
        Omega = self._calc_Omega(W, tau, rho, sigma) if Omega is None else Omega
        dL_dOmega = self._dL_dOmega(z, Omega, chain=True, Omega_inv=Omega_inv)
        tau_prime = self._dL_dtau(z, Omega, W, tau, rho, sigma, dL_dOmega=dL_dOmega)
        rho_prime = self._dL_drho(z, Omega, W, tau, rho, sigma, dL_dOmega=dL_dOmega)
        sigma_prime = self._dL_dsigma(z, Omega, W, tau, rho, sigma, dL_dOmega=dL_dOmega)
        # print(f"{tau_prime[-3:]}, {rho_prime}, {sigma_prime}")
        return -np.r_[tau_prime, rho_prime, sigma_prime]

    def _negloglikelihood_prime_numerical(self, params, z, W, h=1e-6):
        Hs = np.eye(len(params)) * h
        return np.array([(self._negloglikelihood(params+H, z, W) - self._negloglikelihood(params-H, z, W)) / (2*h) for H in Hs])

    def _test_gradient(self, n_channels=6, n_voxels=10, n_trials=7):
        z = np.random.randn(n_voxels, n_trials)
        W = np.random.rand(n_voxels, n_channels)
        tau = np.random.rand(n_voxels)
        rho = 0.5
        sigma = 0.1
        params = np.r_[tau, rho, sigma]
        np.testing.assert_allclose(self._negloglikelihood_prime(params, z, W), 
            self._negloglikelihood_prime_numerical(params, z, W), rtol=1e-6)

    def _test_loglikelihood(self, n_channels=6, n_voxels=10, n_trials=7):
        z = np.random.randn(n_voxels, n_trials)
        W = np.random.rand(n_voxels, n_channels)
        tau = np.random.rand(n_voxels)
        rho = 0.5
        sigma = 0.1
        mv_norm = stats.multivariate_normal(np.zeros(n_voxels), self._calc_Omega(W, tau, rho, sigma))
        np.testing.assert_allclose(self._calc_L(z, W, tau, rho, sigma),
            np.sum(mv_norm.logpdf(z.T)), rtol=1e-6)

    def _calc_Omega(self, W, tau, rho, sigma):
        return (rho + (1-rho)*np.eye(len(tau))) * np.outer(tau, tau) + sigma**2 * W@W.T

    def _calc_L(self, z, W, tau, rho, sigma, Omega=None, Omega_inv=None):
        '''
        L = log(p(b|s; W, Omega))
        z = b - W @ fs
        '''
        Omega = self._calc_Omega(W, tau, rho, sigma) if Omega is None else Omega
        M = math.pinv(Omega) if Omega_inv is None else Omega_inv # Although (4x) slower than inv, pinv is preferred in the numerical world (=inv if invertible and well conditioned)
        n_voxels, n_trials = z.shape
        # For a single sample: -0.5 * (z.T @ M @ z + np.log(np.linalg.det(Omega)) + n_voxels*np.log(2*np.pi))
        # May also use (by Gilles): np.sum(stats.multivariate_normal(np.zeros(n_voxels), Omega).logpdf(z.T))
        # This is less numerically robust: -0.5 * (np.trace(z.T @ M @ z) + n_trials*np.log(np.linalg.det(Omega)) + n_trials*n_voxels*np.log(2*np.pi))
        # return -0.5 * (np.trace(z.T @ M @ z) + n_trials*np.prod(np.linalg.slogdet(Omega)) + n_trials*n_voxels*np.log(2*np.pi))
        return -0.5 * ((z * (M @ z)).sum() + n_trials*np.prod(np.linalg.slogdet(Omega)) + n_trials*n_voxels*np.log(2*np.pi))
        
    def _dL_dOmega(self, z, Omega, chain=False, Omega_inv=None):
        M = math.pinv(Omega) if Omega_inv is None else Omega_inv # Although (4x) slower than inv, pinv is preferred in the numerical world (=inv if invertible and well conditioned)
        n_voxels, n_trials = z.shape
        # deriv = 0.5 * (M.T @ np.outer(z, z) @ M.T - M.T)
        # For a single sample: deriv = 0.5 * (M @ np.outer(z, z) @ M - M) # M.T == M
        deriv = 0.5 * (M @ z @ z.T @ M - n_trials*M)
        if chain:
            return deriv
        else:
            return (2 - np.eye(n_voxels)) * deriv

    def _dL_dtau(self, z, Omega, W, tau, rho, sigma, dL_dOmega=None):
        N = len(tau)
        dL_dOmega = self._dL_dOmega(z, Omega, chain=True) if dL_dOmega is None else dL_dOmega
        # deriv = np.zeros(N)
        # for n in range(N):
        #     e = np.zeros(N)
        #     e[n] = 1
        #     dOmega_dtau_n = (rho+(1-rho)*np.eye(N)) * (np.outer(e, tau) + np.outer(tau, e))
        #     # deriv[n] = np.trace(dL_dOmega @ dOmega_dtau_n)
        #     # The above expression is wasteful, trace(A@B) == sum(A*B.T), and Einstein summation is even faster (but numerically less stable???)
        #     deriv[n] = np.einsum('ij,ji->', dL_dOmega, dOmega_dtau_n)
        # This function is the bottleneck after cProfile.run()
        # https://stackoverflow.com/questions/18854425/what-is-the-best-way-to-compute-the-trace-of-a-matrix-product-in-numpy
        A = dL_dOmega * (rho+(1-rho)*np.eye(N)).T
        # deriv = (A*tau).sum(axis=1) + (A*tau[:,np.newaxis]).sum(axis=0)
        deriv = 2 * (A*tau).sum(axis=1) # A should be symmetric matrix
        return deriv

    def _dL_drho(self, z, Omega, W, tau, rho, sigma, dL_dOmega=None):
        dL_dOmega = self._dL_dOmega(z, Omega, chain=True) if dL_dOmega is None else dL_dOmega
        dOmega_drho = (1 - np.eye(len(tau))) * np.outer(tau, tau)
        # return np.trace(dL_dOmega @ dOmega_drho)
        return (dL_dOmega * dOmega_drho.T).sum()

    def _dL_dsigma(self, z, Omega, W, tau, rho, sigma, dL_dOmega=None):
        dL_dOmega = self._dL_dOmega(z, Omega, chain=True) if dL_dOmega is None else dL_dOmega
        dOmega_dsigma = W@W.T * 2*sigma
        # return np.trace(dL_dOmega @ dOmega_dsigma)
        return (dL_dOmega * dOmega_dsigma.T).sum()


class EnsembleModel(BaseModel):
    def __init__(self, n_ensemble=10, base_model='required', pred_method=None, pred_options=None):
        # Cannot use 1) **kwargs; 2) class as argument. Use instance instead (__class__ + get_params).
        # Otherwise you may get the misleading "TypeError: get_params() missing 1 required positional argument: 'self'".
        # Also cannot modify any argument, otherwise sklearn's clone() method will complain during cross-validation.
        self.n_ensemble = n_ensemble
        self.base_model = base_model
        self.pred_method = pred_method
        self.pred_options = pred_options

    # get_params() is required by sklearn
    def get_params(self, deep=True):
        return dict(n_ensemble=self.n_ensemble, base_model=self.base_model,
            pred_method=self.pred_method, pred_options=self.pred_options)

    def fit(self, X, y):
        # Perform argument validation here (as recommended by sklearn) so that get_params() and __init__() have the same effect 
        # This is refactored so that load() can work without fit()
        self._set_default_params()
        self.models_ = [self.base_model.__class__(**self.base_model.get_params()).fit(X[:,k::self.n_ensemble], y) for k in range(self.n_ensemble)]
        return self # Required by sklearn

    def predict(self, X, method=None, options=None, return_all=False, pred_kws=None):
        method = self.pred_method if method is None else method
        options = self.pred_options if options is None else options
        pred_kws = dict(dict(return_all=(True if method in ['map'] else False)), **({} if pred_kws is None else pred_kws))
        preds = [model.predict(X[:,k::self.n_ensemble], **pred_kws) for k, model in enumerate(self.models_)]
        if method == 'mean':
            y_hat = np.mean([pred[0] if isinstance(pred, tuple) else pred for pred in preds], axis=0)
        elif method == 'map':
            stimulus_domain = pred_kws['stimulus_domain'] if 'stimulus_domain' in pred_kws else self.base_model.get_params()['stimulus_domain']
            posterior = np.mean([prep[options['pidx']] for prep in preds], axis=0)
            y_hat = stimulus_domain[np.argmax(posterior, axis=1)]
        return (y_hat, preds) if return_all else y_hat

    def _set_default_params(self):
        if self.pred_method is None:
            self.pred_method = 'map' if hasattr(self.base_model, '_pidx') else 'mean' 
        self.pred_options = dict(dict(pidx=(self.base_model._pidx if hasattr(self.base_model, '_pidx') else -1)), 
            **({} if self.pred_options is None else self.pred_options))

    def to_dict(self):
        d = super().to_dict()
        d['models_'] = [model.to_dict() for model in d['models_']]
        return d

    def from_dict(self, d):
        self._set_default_params()
        d['models_'] = [self.base_model.__class__(**self.base_model.get_params()).from_dict(model) for model in d['models_']]
        self.__dict__.update(d)
        return self


def shift_distribution(d, stimulus_domain, center_on=None, circular=True):
    '''
    Parameters
    ----------
    d : n_trials * n_domain
    '''
    if center_on is not None:
        tgt_idx = len(stimulus_domain)//2
        src_idx = np.argmin(np.abs(center_on[:,np.newaxis] - stimulus_domain[np.newaxis,:]), axis=-1)
        shift_idx = tgt_idx - src_idx
    shifted = np.zeros_like(d)
    for k, dd in enumerate(d):
        dd = np.roll(dd, shift_idx[k])
        if not circular:
            if shift_idx[k] >= 0:
                dd[:shift_idx[k]] = np.nan
            else:
                dd[shift_idx[k]:] = np.nan
        shifted[k,:] = dd
    return shifted


def discretize_prediction(y_pred, targets, circular_domain=None):
    '''
    Discretize continous prediction to the nearest target.
    Can handle irregular target grid and also circular domain.

    E.g.,
    y_pred = encoding.discretize_prediction(y_hat, arange(8)/8*pi, circular_domain=[0, pi])
    correct = encoding.circular_correct(y_true, y_hat, domain=[0, pi], n_targets=8)
    assert(allclose(mean(y_pred==y_true), mean(correct)))
    '''
    if circular_domain: # Circular domain
        D = circular_domain[-1] - circular_domain[0] # Domain size
        augmented = np.r_[targets, D+targets[0]]
    else: # Non circular domain
        augmented = targets
    idx = np.argmin(np.abs(y_pred[:,np.newaxis] - augmented[np.newaxis,:]), axis=-1) % len(targets)
    return targets[idx]


def circular_correct(y_true, y_pred, domain=None, n_targets=None, tolerance=None, return_dist=False):
    if domain is None:
        domain = [0, 2*np.pi]
    D = domain[-1] - domain[0] # Domain size
    if n_targets is None:
        if tolerance is None:
            raise ValueError('You must provide either "n_targets" or "tolerance".')
        else:
            d = tolerance
    else:
        d = D / n_targets / 2
    dist = np.abs(y_pred - y_true)
    dist = np.minimum(dist, D-dist)
    correct = (dist < d)
    return (correct, dist) if return_dist else correct


if __name__ == '__main__':
    pass



    #!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import os, shutil, ctypes, multiprocessing
from os import path
import parser
from itertools import chain
from scipy import spatial
import numpy as np
from . import six, afni, io, utils, _with_pylab


def map_sequence(seq1, seq2):
    '''
    Generate a mapper from seq1 to seq2, so that seq1[mapper]==seq2,
    i.e, the mapper[k]'s node of seq1 is the k's node of seq2.
    mapper = map_sequence(v2, v1)
    mapper_inv = map_sequence(v1, v2)
    v2m = v2[mapper]
    f2m = mapper_inv[f2]
    assert(all(v2m==v1))
    assert(all(f2m[lexsort(f2m.T)]==f1[lexsort(f1.T)]))
    '''
    assert(seq1.shape[0]==seq2.shape[0])
    indexer1 = np.lexsort(seq1.T) # Don't use `indexer1 = np.argsort(seq1, axis=0)`
    indexer2 = np.lexsort(seq2.T)
    mapper = np.zeros(len(seq1), dtype=int)
    mapper[indexer2] = indexer1
    return mapper


def quadruple_mesh(verts, faces, power=1, mask=None, values=[]):
    '''
    A face will be divided if any of its three nodes are within the mask.
    '''
    if mask is not None:
        mask = set(mask)
    for _ in range(power):
        nv = [v for v in verts]
        nv_parent = {}
        nf = []
        def get_new_vert(n1, n2):
            n_idx = nv_parent.setdefault(tuple(sorted([n1, n2])), len(nv))
            if n_idx == len(nv):
                nv.append((verts[n1]+verts[n2])/2)
            return n_idx
        for f in faces:
            if mask is not None and set(f).isdisjoint(mask):
                # Skip dividing this face
                nf.append(f)
            else: # Divide this face into four new faces
                nv0 = get_new_vert(f[1], f[2])
                nv1 = get_new_vert(f[2], f[0])
                nv2 = get_new_vert(f[0], f[1])
                # The triangle should always list the vertices in a counter-clockwise direction 
                # with respect to an outward pointing surface normal vector [Noah Benson]
                nf.extend([(f[0], nv2, nv1), (nv2, f[1], nv0), (nv1, nv0, f[2]), (nv0, nv1, nv2)])
        verts = nv
        faces = nf
    return np.array(verts), np.array(faces)


def immediate_neighbors(verts, faces, return_array=False):
    '''
    By default, neighbors are represented as a list of sets:
        [set([n00, n01, ...]), set([n10, n11, ...]), ...]
    If return_array=True, a numpy array with a special schema is returned
    to facilitate inter-process memory sharing in multiprocessing:
        [[n_nb0+1, n00, n01, ..., 0, 0], [n_nb1+1, n10, n11, ..., 0, 0], ...] 
    '''
    # neighbors = [np.unique(np.r_[faces[faces[:,0]==idx,1:].ravel(), 
    #     faces[faces[:,1]==idx,::2].ravel(), 
    #     faces[faces[:,2]==idx,:2].ravel()]) for idx in range(verts.shape[0])]
    # The above code is unreasonably slow...
    n_verts = verts.shape[0]
    neighbors = [set() for _ in range(n_verts)] # Don't use [set()]*n_verts
    for face in faces:
        neighbors[face[0]].update(face[1:])
        neighbors[face[1]].update(face[::2])
        neighbors[face[2]].update(face[:2])
    if return_array: # For shared memory parallelism
        n_nb = [len(nbhd) for nbhd in neighbors]
        shared_arr = multiprocessing.Array(ctypes.c_int32, n_verts*(max(n_nb)+1), lock=False)
        arr = np.frombuffer(shared_arr, dtype=np.int32).reshape(n_verts,-1)
        for idx in range(n_verts):
            arr[idx,0] = n_nb[idx] + 1
            arr[idx,1:arr[idx,0]] = list(neighbors[idx])
        neighbors = shared_arr
    return neighbors


def interp_over_mesh(verts, faces, indices, values, radius=3, neighbors=None, n_verts=None):
    '''
    By reusing precomputed neighbors (multiprocessing.Array) and n_verts, 
    one can use inter-process memory sharing to efficiently interpolate multiple dsets in parallel.
    verts and faces are unused in this case.
    '''
    if neighbors is None:
        neighbors = immediate_neighbors(verts, faces)
    if n_verts is None:
        n_verts = verts.shape[0]
    indices_set = set(indices)
    new_val = np.zeros(n_verts)
    if isinstance(neighbors, ctypes.Array): # For shared memory parallelism
        neighbors = np.frombuffer(neighbors, dtype=np.int32).reshape(n_verts,-1)
        for idx in range(n_verts):
            nbhd = set(neighbors[idx,1:neighbors[idx,0]])
            new_nbhd = nbhd
            for _ in range(1, radius):
                ext_nbhd = set()
                for nb in new_nbhd:
                    ext_nbhd.update(neighbors[nb,1:neighbors[nb,0]])
                new_nbhd = ext_nbhd.difference(nbhd)
                nbhd.update(ext_nbhd)
            new_val[idx] = np.mean([values[nb] for nb in nbhd if nb in indices_set])
    else: # neighbors is a list of sets
        for idx in range(n_verts):
            nbhd = neighbors[idx].copy()
            new_nbhd = nbhd
            for _ in range(1, radius):
                ext_nbhd = set()
                for nb in new_nbhd:
                    ext_nbhd.update(neighbors[nb])
                new_nbhd = ext_nbhd.difference(nbhd)
                nbhd.update(ext_nbhd)
            # new_val[idx] = np.mean(values[np.isin(indices, list(nbhd))])
            # The above code is slow again...
            new_val[idx] = np.mean([values[nb] for nb in nbhd if nb in indices_set])
    return new_val


def interp_dset(fdset, fmesh, prefix):
    '''
    `fmesh` can be either:
    - fname for a high density surface mesh
    - (neighbors, n_verts) for using shared memory parallelism
    '''
    indices, values = io.read_niml_bin_nodes(fdset)
    if isinstance(fmesh, six.string_types):
        verts, faces = io.read_surf_mesh(fmesh)
        new_val = interp_over_mesh(verts, faces, indices, values)
    else: # fmesh = (neighbors, n_verts), for shared memory parallelism
        new_val = interp_over_mesh(None, None, indices, values, neighbors=fmesh[0], n_verts=fmesh[1])
    io.write_niml_bin_nodes(utils.fname_with_ext(prefix, '.niml.dset'), np.arange(new_val.shape[0]), new_val)


def compute_faces_norm(verts, faces, F):
    e01 = verts[F[:,1],:] - verts[F[:,0],:]
    e12 = verts[F[:,2],:] - verts[F[:,1],:]
    e01 /= np.linalg.norm(e01, axis=1, keepdims=True)
    e12 /= np.linalg.norm(e12, axis=1, keepdims=True)
    norms = np.cross(e01, e12, axis=1)
    norms /= np.linalg.norm(norms, axis=1, keepdims=True)
    return norms


def surrouding_faces(verts, faces, v):
    return faces[np.any(faces == v, axis=1),:]


def compute_verts_norm(verts, faces, V):
    norms = np.zeros([V.shape[0],3], dtype=verts.dtype)
    for k, v in enumerate(V):
        F = surrouding_faces(verts, faces, v)
        n = compute_faces_norm(verts, faces, F)
        n = np.sum(n, axis=0)
        n /= np.linalg.norm(n)
        norms[k,:] = n
    return norms
    

def compute_verts_area(verts, faces, dtype=None):
    '''
    Compute element area for each vertex.
    '''
    # Compute area for each face triangle
    A = verts[faces[:,0],:]
    B = verts[faces[:,1],:]
    C = verts[faces[:,2],:]
    face_areas = 0.5*np.linalg.norm(np.cross(B-A, C-A), axis=1)
    vert_areas = np.zeros(verts.shape[0], dtype=dtype)
    # Attribtue the face area to each of its three vertices
    # for k, f in enumerate(faces):
    #     vert_areas[f] += face_areas[k]
    # vert_areas /= 3.0
    # The above code with advanced indexing is 2x slower...
    for (a, b, c), x in zip(faces, face_areas/3.0):
        vert_areas[a] += x
        vert_areas[b] += x
        vert_areas[c] += x
    return vert_areas


def smooth_verts_data(verts, faces, data, factor=0.1, n_iters=1, dtype=None):
    for _ in range(n_iters):
        smooth_data = np.zeros(len(data), dtype=dtype)
        counts = np.zeros(len(data), dtype=dtype)
        for a, b, c in faces:
            smooth_data[a] += data[b] + data[c]
            smooth_data[b] += data[c] + data[a]
            smooth_data[c] += data[a] + data[b]
            counts[a] += 2
            counts[b] += 2
            counts[c] += 2
        smooth_data = factor * smooth_data/counts + (1-factor) * data
        data = smooth_data
    return smooth_data


def compute_intermediate_mesh(inner, outer, alpha, method='equivolume', dtype=None):
    alpha = np.array(alpha, dtype=dtype).reshape(-1, 1)
    vin, fin = io.read_surf_mesh(inner, dtype=dtype) if isinstance(inner, six.string_types) else inner
    vout, fout = io.read_surf_mesh(outer, dtype=dtype) if isinstance(outer, six.string_types) else outer
    Ain = compute_verts_area(vin, fin, dtype=dtype)
    Aout = compute_verts_area(vout, fout, dtype=dtype)
    if method in ['equivolume', 'equivolume_inside']:
        smooth_factor = 0.1
        smooth_iters = 2
        Ain = smooth_verts_data(vin, fin, Ain, factor=smooth_factor, n_iters=smooth_iters, dtype=dtype)
        Aout = smooth_verts_data(vout, fout, Aout, factor=smooth_factor, n_iters=smooth_iters, dtype=dtype)
        if method == 'equivolume':
            rho = 1 / (Aout - Ain) * (-Ain + np.sqrt(alpha * Aout**2 + (1-alpha) * Ain**2))
        elif method == 'equivolume_inside':
            rho = alpha + np.zeros(len(Ain), dtype=dtype)
            inside = ((0 <= alpha) & (alpha <= 1)).ravel()
            rho[inside,:] = 1 / (Aout - Ain) * (-Ain + np.sqrt(alpha[inside,:] * Aout**2 + (1-alpha[inside,:]) * Ain**2))
    elif method == 'equidistance':
        rho = alpha
    verts = (1-rho[...,np.newaxis]) * vin + rho[...,np.newaxis] * vout
    faces = fin
    return verts.squeeze(), faces


def compute_voxel_depth(xyz, inner, outer, S2E_mat, method='equivolume', n_jobs=4, dtype=None, lock=None):
    '''
    Parameters
    ----------
    method : str
        "equivolume"
        "equidistance"

    Notes
    -----
    1. Unfortunately, dtype=np.float32 doesn't work for high density surface meshes
       (because the element area becomes zero in some locations, which is invalid). 
       Although it does work for ordinary meshes, it is unnecessary in that case.
    '''
    if isinstance(xyz, six.string_types):
        xyz = io.Mask(xyz, kind='full').xyz
    xyz = xyz.astype(dtype)
    if isinstance(S2E_mat, six.string_types):
        S2E_mat = afni.get_S2E_mat(S2E_mat, mat='S2B')
    S2E_mat = S2E_mat.astype(dtype)
    if method == 'equivolume':
        method = 'equivolume_inside'
    if lock is None:
        lock = multiprocessing.Lock()
    min_depth, max_depth = -0.2, 1.2
    n_depths = round((max_depth - min_depth)/0.1) + 1
    alphas = np.linspace(min_depth, max_depth, n_depths, dtype=dtype)
    print('>> Compute intermediate meshes...')
    verts, faces = compute_intermediate_mesh(inner, outer, alphas, method=method, dtype=dtype) # 234s
    n_faces = faces.shape[0]
    LPI2RAI = np.array([-1, -1, 1], dtype=dtype)
    verts = np.dot(S2E_mat[:,:3], (verts*LPI2RAI).transpose(0,2,1)).transpose(1,2,0) + S2E_mat[:,3]
    face_xyz = (verts[:,faces[:,0],:] + verts[:,faces[:,1],:] + verts[:,faces[:,2],:]).reshape(-1,3) / 3
    print('>> Construct k-d tree...')
    # kdt = spatial.cKDTree(face_xyz.reshape(-1,3)) # 171s
    kdt = spatial.cKDTree(face_xyz) # This is slightly more memory efficient
    depths = utils.SharedMemoryArray.zeros(xyz.shape[0], dtype=dtype, lock=False) # np.zeros(xyz.shape[0])
    print('>> Compute cortical depth...')
    def compute_depth(ids, depths, xyz, kdt, verts, faces, alphas, n_faces, n_depths, min_depth, max_depth):
        for k in ids:
            p = xyz[k]
            # idx = np.argmin(np.linalg.norm(p - face_xyz, axis=-1)) // faces.shape[0]
            idx = kdt.query(p)[1] # This is like 4000x faster!
            fidx = idx % n_faces
            didx = idx // n_faces
            if didx == 0:
                depths[k] = min_depth
            elif didx == n_depths-1:
                depths[k] = max_depth
            else:
                A, B, C = verts[didx-1:didx+2][:,faces[fidx,:],:].swapaxes(0,1)
                N = np.cross(B - A, C - A)
                N = N / np.linalg.norm(N, axis=-1, keepdims=True)
                T = np.sum(A*N, axis=-1) - np.sum(p*N, axis=-1)
                W = np.abs(T)
                if T[0]*T[1] < 0:
                    w = W[1] / (W[0] + W[1])
                    depths[k] = w * alphas[didx-1] + (1-w) * alphas[didx]
                else:
                    w = W[1] / (W[2] + W[1])
                    depths[k] = w * alphas[didx+1] + (1-w) * alphas[didx]
    with lock:
        pc = utils.PooledCaller(pool_size=n_jobs)
        pc(pc.run(compute_depth, ids, depths, xyz, kdt, verts, faces, alphas, n_faces, n_depths, min_depth, max_depth) 
            for ids in pc.idss(len(depths)))
    return depths


def create_lamina_mesh(fname, inner, outer, alpha, method='equivolume'):
    if method == 'equivolume':
        method = 'equivolume_inside'
    verts, faces = compute_intermediate_mesh(inner, outer, alpha, method=method)
    io.write_surf_mesh(fname, verts, faces)


def transform_mesh(transform, in_file, out_file):
    '''
    Parameters
    ----------
    transform : mripy.preprocess.Transform object
    in_file, out_file : surface mesh file name (*.gii or *.asc)
    '''
    verts, faces = io.read_surf_mesh(in_file)
    verts = transform.apply_to_xyz(verts, convention='NIFTI')
    io.write_surf_mesh(out_file, verts, faces)


def transform_suma(transform, suma_dir, out_dir=None):
    '''
    BUG: Only part of the most used surfaces and volumes are transformed.
    '''
    if out_dir is None:
        out_dir = suma_dir + '.al'
    pc = utils.PooledCaller()
    ext = afni.get_surf_type(suma_dir)
    # Copy SUMA files
    print('>> Copy SUMA dir...')
    shutil.copytree(suma_dir, out_dir)
    # Transform anatomical surface meshes
    for surf in ['pial', 'smoothwm', 'white']:
        for hemi in ['lh', 'rh']:
            surf_file = f"{out_dir}/{hemi}.{surf}{ext}"
            if path.exists(surf_file):
                pc.run(transform_mesh, transform, surf_file, surf_file)
    # Transform useful volumes
    surf_vol = afni.get_surf_vol(out_dir)
    pc.run(transform.apply, surf_vol, surf_vol)
    # Transform benson14 volumes if exists
    for dset in ['varea', 'eccen', 'angle', 'sigma']:
        f = f"{out_dir}/benson14_{dset}.nii.gz"
        if path.exists(f):
            pc.run(transform.apply, f, f, interp='NearestNeighbor' if dset=='varea' else None)
    pc.wait()


def dset2roi(f_dset, f_roi=None, colors=None):
    '''
    ROI number starts from 1 (nodes with data==0 are ignored).
    See also: FSread_annot
    '''
    if f_roi is None:
        f_roi = '.'.join(f_dset.split('.')[:-2] + ['1D', 'roi'])
    nodes, data = io.read_niml_bin_nodes(f_dset)
    nodes = nodes[data!=0]
    data = data[data!=0].astype(int)
    if colors is None:
        np.savetxt(f_roi, np.c_[nodes, data], fmt='%d')
    else:
        np.savetxt(f_roi, np.c_[nodes, data, colors[(data-1)%len(colors)]], 
            fmt=['%d', '%d', '%.6f', '%.6f', '%.6f'])


def _surface_calc(expr=None, out_file=None, **kwargs):
    '''
    Different input dsets are allowed to have different nodes coverage.
    Only values on shared nodes are returned or written.
    '''
    used = parser.expr(expr).compile().co_names # Parse variables in `expr`
    variables = {}
    shared_nodes = None
    for var, fname in kwargs.items():
        if var in used: # Only consider used variables in `expr`
            nodes, values = io.read_surf_data(fname)
            variables[var] = (nodes, values)
            if shared_nodes is None:
                shared_nodes = set(nodes)
            else:
                shared_nodes = shared_nodes.intersection(nodes)
    shared_nodes = np.array(sorted(shared_nodes)) # `sorted` is required
    for var, (nodes, values) in variables.items():
        shared = np.in1d(nodes, shared_nodes)
        # A test for surface dataset compatibility (duck typing)
        # TODO: There could be more direct check by looking at the header of the dataset
        assert(np.all(nodes[shared]==shared_nodes)) 
        variables[var] = values[shared]
    v = _with_pylab.pylab_eval(expr, **variables)
    if out_file is not None:
        io.write_surf_data(out_file, shared_nodes, v)
    return shared_nodes, v


def surface_calc(expr=None, out_file=None, **kwargs):
    out_file = afni.infer_surf_dset_variants(out_file)
    kwargs = {k: afni.infer_surf_dset_variants(v) for k, v in kwargs.items()}
    for hemi in ['lh', 'rh']:
        _surface_calc(expr=expr, out_file=out_file[hemi], **{k: v[hemi] for k, v in kwargs.items()})


class SurfMask(object):
    def __init__(self, master):
        if master is not None:
            nodes, values = io.read_surf_data(master)
            self.nodes = nodes[values > 0]

    @classmethod
    def from_expr(cls, expr=None, **kwargs):
        mask = cls(master=None)
        nodes, values = _surface_calc(expr=expr, **kwargs)
        mask.nodes = nodes[values > 0]
        return mask

    def __repr__(self):
        return f"{self.__class__.__name__} ({len(self.nodes)} nodes)"

    def dump(self, fname):
        if isinstance(fname, str):
            fname = [fname]
        values = []
        for f in fname:
            n, v = io.read_surf_data(f)
            assert(np.all(np.in1d(self.nodes, n))) # SurfMask nodes must be covered by dset nodes
            shared = np.in1d(n, self.nodes)
            assert(np.all(n[shared]==self.nodes)) # Make sure nodes order are correspondent
            values.append(v[shared])
        return np.array(values).T.squeeze()


class Surface(object):
    def __init__(self, suma_dir, surf_vol=None):
        self.suma_dir = suma_dir
        self.surf_vol = 'SurfVol_Alnd_Exp.nii' if surf_vol is None else surf_vol
        if not path.exists(self.surf_vol):
            raise ValueError(f'>> Cannot find "{self.surf_vol}" on current directory\n"{os.getcwd()}"')
        self.specs = afni.get_suma_spec(self.suma_dir)
        self.subj = afni.get_suma_subj(self.suma_dir)
        self.surfs = ['pial', 'smoothwm', 'inflated', 'sphere.reg']
        self.hemis = ['lh', 'rh']
        self.surf_ext = afni.get_surf_type(self.suma_dir)
        self.info = {hemi: io.read_surf_info(f"{self.suma_dir}/{hemi}.inflated{self.surf_ext}") for hemi in self.hemis}

    def __repr__(self):
        n_verts = f'{self.info["lh"]["n_verts"]}/{self.info["rh"]["n_verts"]} verts'
        n_faces = f'{self.info["lh"]["n_faces"]}/{self.info["rh"]["n_faces"]} faces'
        return f'<Surface  | {n_verts}, {n_faces}, suma_dir="{self.suma_dir}", \n surf_vol="{self.surf_vol}", surf_ext="{self.surf_ext}">'

    def _get_surf2exp_transform(self, exp_anat):
        pass

    def _get_spherical_coordinates(self, hemi, symmetric=True):
        verts = io.read_surf_mesh(path.join(self.suma_dir, f"{hemi}.sphere.reg{self.surf_ext}"))[0]
        # verts[:,2] sometimes can be greater than 100 or less than -100
        theta = np.arccos(np.maximum(np.minimum(verts[:,2], 100), -100)/100) # Polar (inclination) angle, [0,pi] 
        phi = np.arctan2(verts[:,1], verts[:,0]) # Azimuth angle, (-pi,pi]
        if symmetric and hemi == 'rh':
            phi = -phi + np.pi
            phi[phi>np.pi] -= 2*np.pi
        return theta, phi

    def to_1D_dset(self, prefix, node_values):
        np.savetxt(f"{prefix}.1D.dset", np.c_[np.arange(len(node_values)), node_values], fmt='%.6f')

    def vol2surf(self, in_file, out_file, func='median', depth_range=[0,1], mask_file=None, truncate=True):
        '''
        Parameters
        ----------
        in_file : str, 
            E.g., "beta.nii" or "stats.loc_REML.nii'[L-R#0_Tstat]'".
        out_file : str
            "beta.niml.dset" will automatically generate ["lh.beta.niml.dset", "rh.beta.niml.dset"].
        func : str
            ave, median, max, midpoint, etc.
        mask_file : str
            Volume mask file.
        truncate : bool
            If True, vertices whose value is zero will be omitted in the output surface dataset.
            `surface_calc()` will handle such (partial) surface dataset correctly.
            But if you need to use `3dcalc`, set truncate=False and output all vertices.

        About "-f_index nodes"
        ----------------------
        [1] https://afni.nimh.nih.gov/pub/dist/doc/program_help/3dVol2Surf.html
            When taking the average along one node pair segment using 10 node steps,
            perhaps 3 of those nodes may occupy one particular voxel.  In this case, 
            does the user want the voxel counted only once (-f_index voxels), 
            or 3 times (-f_index nodes)?  Each way makes sense.
        '''
        mask_cmd = f"-cmask {mask_file}" if mask_file is not None else ''
        truncate_cmd = f"-oob_value 0 {'-oom_value 0' if mask_file is not None else ''}" if not truncate else ''
        out_dir, prefix, ext = afni.split_out_file(out_file, split_path=True, trailing_slash=True)
        output_1D = '.1D' in ext
        pc = utils.PooledCaller()
        for hemi in self.hemis:
            fo_niml = f"{out_dir}{hemi}.{prefix}.niml.dset"
            fo_1D = f"{out_dir}{hemi}.{prefix}.1D.dset"
            out_1D_cmd = f"-out_1D {fo_1D}" if output_1D else ''
            pc.run(f"3dVol2Surf \
                -spec {self.specs[hemi]} \
                -surf_A smoothwm \
                -surf_B pial \
                -sv {self.surf_vol} \
                -grid_parent {in_file} \
                {mask_cmd} {truncate_cmd} \
                -map_func {func} \
                -f_steps 20 -f_index nodes \
                -f_p1_fr {depth_range[0]} -f_pn_fr {depth_range[1]-1} \
                -out_niml {fo_niml} {out_1D_cmd} -overwrite", 
                _error_pattern='error', _suppress_warning=True)
        pc.wait()
        return pc._log

    def surf2vol(self, base_file, in_files, out_file, func='median', combine='mean', depth_range=[0,1], mask_file=None):
        '''
        Parameters
        ----------
        in_files : str, list, or dict
            "beta.niml.dset" will be expanded as ["lh.beta.niml.dset", "rh.beta.niml.dset"],
            whereas "lh.beta.niml.dset" will be treated as is.
        func : str
            ave, median, mask2, count, etc.
        combine : str
            l+r, max(l,r), consistent, mean, lh, rh, etc.
        mask_file : str
            Volume mask file.

        About "-f_index nodes"
        ----------------------
        [1] https://afni.nimh.nih.gov/pub/dist/doc/program_help/3dSurf2Vol.html
            Current setting may be preferred if the user wishes to have the average 
            weighted by the number of points occupying a voxel (-f_index nodes), 
            not just the number of node pair segments (-f_index voxels).
        '''
        in_files = afni.infer_surf_dset_variants(in_files, hemis=self.hemis)
        mask_cmd = f"-cmask {mask_file}" if mask_file is not None else ''
        temp_dir = utils.temp_folder()
        pc = utils.PooledCaller()
        for hemi, dset in in_files.items():
            prefix, ext = afni.split_out_file(dset)
            input_cmd = f"-sdata_1D {dset}" if '.1D' in ext else f"-sdata {dset}"
            pc.run(f"3dSurf2Vol \
                -spec {self.specs[hemi]} \
                -surf_A smoothwm \
                -surf_B pial \
                -sv {self.surf_vol} \
                -grid_parent {base_file} \
                {input_cmd} {mask_cmd} \
                -map_func {func} \
                -datum float \
                -f_steps 20 -f_index nodes \
                -f_p1_fr {depth_range[0]} -f_pn_fr {depth_range[1]-1} \
                -prefix {temp_dir}/{hemi}.nii -overwrite", 
                _error_pattern='error', _suppress_warning=True)
        pc.wait()
        if len(in_files) > 1:
            if combine == 'consistent':
                combine_expr = 'notzero(l)*iszero(r)*l+iszero(l)*notzero(r)*r+notzero(l)*notzero(r)*equals(l,r)*l'
            elif combine == 'mean':
                combine_expr = 'notzero(l)*iszero(r)*l+iszero(l)*notzero(r)*r+notzero(l)*notzero(r)*(l+r)/2'
            elif combine == 'lh':
                combine_expr = 'l+iszero(l)*notzero(r)*r'
            elif combine == 'rh':
                combine_expr = 'r+iszero(r)*notzero(l)*l'
            else:
                combine_expr = combine
            pc.run1(f"3dcalc -l {temp_dir}/lh.nii -r {temp_dir}/rh.nii \
                -expr '{combine_expr}' -prefix {out_file} -overwrite")
        else:
            pc.run1(f"3dcopy {temp_dir}/{hemi}.nii {out_file} -overwrite")
        shutil.rmtree(temp_dir)
        return pc._log

    def mask_ribbon(self, in_file, out_file, depth_file=None):
        temp_file = utils.temp_prefix(suffix='.niml.dset')
        pc = utils.PooledCaller()
        _log = self.vol2surf(in_file, temp_file, func='max', depth_range=[0.2, 0.8])
        _log += self.surf2vol(in_file, temp_file, out_file, func='mode') # Round at border
        for hemi in self.hemis:
            os.remove(f"{hemi}.{temp_file}")
        if depth_file is not None:
            pc.run1(f"3dcalc -a {out_file} -d {depth_file} -expr 'a*step(d)*step(1-d)' \
                -prefix {out_file} -overwrite")
        return _log + pc._log


if __name__ == '__main__':
    pass
